;; -----------------------------------------------------------------------
 ; Indiana University Opie Lab Software License, Version 1.2
 ;
 ; Copyright (C) 2002 The Trustees of Indiana University.
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without
 ; modification, are permitted provided that the following conditions are
 ; met:
 ;
 ; 1) All redistributions of source code must retain the above
 ;    copyright notice, the list of authors in the original source
 ;    code, this list of conditions and the disclaimer listed in this
 ;    license;
 ;
 ; 2) All redistributions in binary form must reproduce the above
 ;    copyright notice, this list of conditions and the disclaimer
 ;    listed in this license in the documentation and/or other
 ;    materials provided with the distribution;
 ;
 ; 3) Any documentation included with all redistributions must include
 ;    the following acknowledgement:
 ;
 ;      "This product includes software developed by the Indiana
 ;      University Opie Lab.  For further information please visit
 ;      http://www.cs.indiana.edu/~dswise"
 ;
 ;    Alternatively, this acknowledgment may appear in the software
 ;    itself, and wherever such third-party acknowledgments normally
 ;    appear.
 ;
 ; 4) The name "Indiana University" or "Indiana University
 ;    Opie Lab" shall not be used to endorse or promote
 ;    products derived from this software without prior written
 ;    permission from Indiana University.  For written permission,
 ;    please contact http://www.cs.indiana.edu/~dswise.
 ;
 ; 5) Products derived from this software may not use "Indiana
 ;    Univeristy" name nor may "Indiana University" appear in their name,
 ;    without prior written permission of the Indiana University.
 ;
 ; Indiana University provides no reassurances that the source code
 ; provided does not infringe the patent or any other intellectual
 ; property rights of any other entity.  Indiana University disclaims any
 ; liability to any recipient for claims brought by any other entity
 ; based on infringement of intellectual property rights or otherwise.
 ;
 ; LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH
 ; NO WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA
 ; UNIVERSITY GIVES NO WARRANTIES AND MAKES NO REPRESENTATION THAT
 ; SOFTWARE IS FREE OF INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR
 ; OTHER PROPRIETARY RIGHTS.  INDIANA UNIVERSITY MAKES NO WARRANTIES THAT
 ; SOFTWARE IS FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP
 ; DOORS", "WORMS", OR OTHER HARMFUL CODE.  LICENSEE ASSUMES THE ENTIRE
 ; RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR ASSOCIATED MATERIALS,
 ; AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION GENERATED USING
 ; SOFTWARE.
;; ------------------------------------------------------------------------

; scil-parse.ss
; created by Brooke Chenoweth

; Take Steve's lists and put them into records.  Build a tree out of the records

; some of this has been changed from the specifications to match the perl

; As far as I know, this file matches the perl code, except for
; constants, which Steve is planning to change. Also, I don't link up
; pragmas in the tree.

; scil-records.ss must be loaded before this file
; call (build-tree-from-file filename) to build the scil tree

;(load "scil-records.ss")  

(case-sensitive #t) ; since C is case sensitive, we'd better make
		    ; Scheme case sensitive as well

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;; stuff to build the records from the lists 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define read-scil-list
  (lambda (filename)
    (with-input-from-file filename
      (lambda ()
	(let ([scil-list (read)])
	  (make-scil_structure
	   (parse-il_header (car scil-list))
	   (parse-memory_region (cadr scil-list))
	   (map parse-memory_region (cddr scil-list))))))))


; I've decided to sort lists to simplify lookup when linking
(define parse-memory_region
  (lambda (ls)
    (make-memory_region
     (car ls)                                        ; number 
     (cadr ls)					     ; name   
     (map parse-source_file (sort-by-car (caddr ls)))		     ; source_file_list   
     (map parse-scope (sort-by-car (list-ref ls 3)))		     ; scope_list         
     (map parse-pragma (sort-by-car (list-ref ls 4)))		     ; pragma_list        
     (map parse-type (sort-by-car (list-ref ls 5)))		     ; type_list          
     (map parse-variable (sort-by-car (list-ref ls 6)))	     ; variable_list      
     (map parse-routine (sort-by-car (list-ref ls 7)))	     ; routine_list       
     (map parse-field (sort-by-car (list-ref ls 8)))		     ; field_list         
     (map parse-dynamic_init (remove '() (list-ref ls 9)))           ; dynamic_init_list  
     (map parse-param_type (sort-by-car (list-ref ls 10)))	     ; param_type_list    
     (map parse-constant (sort-by-car (list-ref ls 11)))           ; constant_list      
     (map parse-statement (sort-by-car (list-ref ls 12)))	     ; statement_list     
     (map parse-expression (sort-by-car (list-ref ls 13)))           ; expression_list
     (map parse-label (sort-by-car (list-ref ls 14)))
     '()
     '()
     '()
     '())))

; take a list of lists and sort them in increasing order by the number
; in the car of each list
(define sort-by-car
  (let ([pred (lambda (ls1 ls2)
		(< (car ls1) (car ls2)))])
    (lambda (ls)
      (sort pred ls))))
     
  
(define parse-il_header
  (lambda (ls)
    (make-il_header
     (parse-address (car ls))                   ;primary_source_file             
     (parse-address (cadr ls))  	        ;primary_scope                   
     (parse-address (caddr ls))	                ;main_routine                    
     (list-ref ls 3)			        ;compiler_version                
     (list-ref ls 4)			        ;time_of_compilation             
     (convert-bool (list-ref ls 5))    	        ;plain_chars_are_signed          
     (list-ref ls 6)			        ;source_language                 
     (convert-bool (list-ref ls 7))	        ;pcc_compatability_mode          
     (convert-bool (list-ref ls 8))    	        ;enum_type_is_integral           
     (list-ref ls 9)			        ;default_max_member_alignment  
     (convert-bool (list-ref ls 10))   	        ;near_and_far_are_enabled                     
     (convert-bool (list-ref ls 11))	        ;far_data_pointers                            
     (convert-bool (list-ref ls 12))	        ;far_code_pointers                            
     (convert-bool (list-ref ls 13))	        ;UCN_identifiers_used                         
     (parse-address (list-ref ls 14)))))        ;nontag_types_used_in_exception_or_rtti 

(define parse-source_file
  (lambda (ls)
    (make-source_file
     (car ls)                                ; number                    
     (parse-address (cadr ls))		     ; next                      
     (caddr ls)				     ; file_name                 
     (list-ref ls 3)			     ; full_name                 
     (list-ref ls 4)			     ; name_as_written           
     (list-ref ls 5)			     ; first_seq_number          
     (list-ref ls 6)			     ; last_seq_number           
     (list-ref ls 7)			     ; first_line_number         
     (parse-address (list-ref ls 8))	     ; first_child_file          
     (parse-address (list-ref ls 9))	     ; last_child_file           
     (convert-bool (list-ref ls 10))	     ; related_file_implicit_include_done 
     (convert-bool (list-ref ls 11))	     ; is_include_file                    
     (convert-bool (list-ref ls 12))	     ; included_by_system_include         
     (convert-bool (list-ref ls 13))	     ; included_by_preinclude             
     (convert-bool (list-ref ls 14)))))	     ; from_system_include_dir            

(define parse-scope
  (lambda (ls)
    (make-scope
     (car ls)                          ; number    
     (parse-address (cadr ls))         ; next      
     (caddr ls) 		       ; kind      
     (parse-address (list-ref ls 3))   ; assoc_block	
     (parse-address (list-ref ls 4))   ; lifetime		
     (parse-address (list-ref ls 5))   ; constants		
     (parse-address (list-ref ls 6))   ; types		
     (parse-address (list-ref ls 7))   ; variables		
     (parse-address (list-ref ls 8))   ; nonstatic_variables   
     (parse-address (list-ref ls 9))   ; labels		
     (parse-address (list-ref ls 10))  ; routines		
     (parse-address (list-ref ls 11))  ; asm_entries	
     (parse-address (list-ref ls 12))  ; scopes		
     (parse-address (list-ref ls 13))  ; dynamic_init	
     (parse-address (list-ref ls 14))  ; pragmas		
     (parse-scope_tail (caddr ls) (list-tail ls 15))))) ; scope_tail

(define parse-scope_tail
  (lambda (kind ls)
    (case kind
      [(sck_func_prototype sck_condition) (make-func_prototype_scope_tail
					    (parse-address (car ls)))] ; assoc_type
      [(sck_function) (make-function_scope_tail
			(parse-address (car ls))       ; routine_ptr
			(parse-address (cadr ls))      ; parameters
			(parse-address (caddr ls)))]   ; return_value_variable
      [(sck_class_struct_union) (make-class_struct_union_scope_tail
				 (parse-address (car ls)))] ; using_decls
      [else '()])))  ; this may not be the best behavior, but it seems reasonable to me

(define parse-pragma
  (lambda (ls)
    '()))
;     (make-pragma
;      (car ls)                        ; number
;      (parse-address (cadr ls))       ; next
;      (parse-address (caddr ls))      ; entity
;      (list-ref ls 3)		     ; position
;      (list-ref ls 4)                 ; pragma_text
;      (convert-bool (list-ref ls 5))  ; ignore_in_backend 
;      (list-ref ls 6))))              ; kind	

(define parse-type
  (lambda (ls)
    (make-type
     (car ls)                                             ; number            
     (parse-address (cadr ls))		     	          ; next              
     (caddr ls)						  ; kind              
     (list-ref ls 3)					  ; size              
     (list-ref ls 4)					  ; alignment         
     (convert-bool (list-ref ls 5))			  ; used_in_exception_or_rtti 
     (convert-bool (list-ref ls 6))			  ; declared_in_function_prototype
     (convert-bool (list-ref ls 7))			  ; is_builtin_va_list 
     (parse-address (list-ref ls 8))		          ; typeinfo_var       
     (list-ref ls 9)					  ; name               
     (list-ref ls 10)                                     ; unmangled_name
     (list-ref ls 11)					  ; decl_position_seq  
     (list-ref ls 12)					  ; decl_position_column      
     (convert-bool (list-ref ls 13))			  ; referenced          
     (convert-bool (list-ref ls 14))			  ; needed             
     (convert-bool (list-ref ls 15))                      ; is_local_to_function
     (list-ref ls 16)					  ; name_linkage              
     (convert-bool (list-ref ls 17))                      ; has_associated_pragma
     (convert-bool (list-ref ls 18))                      ; static_used_by_instantiation
     (parse-address (list-ref ls 19))		          ; reference                 
     (parse-address (list-ref ls 20))		          ; pointer                   
     (parse-address (list-ref ls 21))		          ; qualified                 
     (parse-type_tail (caddr ls) (list-tail ls 22)))))	  ; type_tail 

(define parse-type_tail
  (lambda (kind ls)
    (case kind
      [(tk_integer) (make-int_type_tail
		     (car ls)                               ; int_kind                   
		     (convert-bool (cadr ls))		    ; explicitly_signed         
		     (parse-address (caddr ls))	            ; enum_info_constant_list   
		     (parse-address (list-ref ls 3))        ; enum_info_affiliated_type 
		     (convert-bool (list-ref ls 4)))]	    ; bool_type                 
      [(tk_float) (make-float_type_tail
		   (car ls))]        ; float_kind
      [(tk_pointer) (make-pointer_type_tail
		     (parse-address (car ls))       ; type_pointed_to       
		     (convert-bool (cadr ls)))]	    ; is_reference
      [(tk_routine) (make-routine_type_tail
		     (parse-address (car ls))                   ; return_type                   
		     (parse-address (cadr ls))		        ; param_type_list               
		     (parse-address (caddr ls))		        ; assoc_routine                 
		     (convert-bool (list-ref ls 3))		; has_ellipsis                  
		     (convert-bool (list-ref ls 4)) 		; prototyped                    
		     (list-ref ls 5)				; routine_name_linkage          
		     (convert-bool (list-ref ls 6))		; routine_name_linkage_is_explicit
		     (convert-bool (list-ref ls 7))		; lint_argsused_flag            
		     (convert-bool (list-ref ls 8))		; value_returned_by_cctor       
		     (convert-bool (list-ref ls 9))		; assoc_routine_is_ctor         
		     (convert-bool (list-ref ls 10))		; assoc_routine_is_dtor         
		     (list-ref ls 11)				; lint_varargs_count           
		     (list-ref ls 12)				; arg_pragma                   
		     (list-ref ls 13)				; qualifiers                   
		     (parse-address (list-ref ls 14)))]         ; prototype_scope		     	                      
      [(tk_array) (make-array_type_tail                      
		   (parse-address (car ls))                     ; element_type            
		   (convert-bool (cadr ls))			; is_static               
		   (convert-bool (caddr ls))			; is_variable_size_array  
		   (convert-bool (list-ref ls 3))		; is_vla                  
		   (convert-bool (list-ref ls 4))		; has_assoc_vla_dimension 
		   (parse-address (list-ref ls 5))		; element_count_expr      
		   (list-ref ls 6)				; number_of_elements      
		   (parse-address (list-ref ls 7)))]	        ; bound_constant          
      [(tk_struct tk_union) (make-struct_type_tail
			     (parse-address (car ls))        ; field_list             
			     (convert-bool (cadr ls))	     ; any_const_member       
			     (convert-bool (caddr ls))       ; any_mutable_member     
			     (convert-bool (list-ref ls 3))  ; abstract               
			     (convert-bool (list-ref ls 4))  ; definition_needed      
			     (list-ref ls 5))]		     ; max_member_alignment   
      [(tk_typeref) (make-typeref_type_tail
		     (parse-address (car ls))       ; type_ref_type   
		     (cadr ls)			    ; qualifiers      
		     (convert-bool (caddr ls)))]    ; explicit_memory_attribute_made_implicit
      [(tk_ptr_to_member) (make-ptr_to_member_type_tail
			   (parse-address (car ls)))] ; class_of_which_a_member
      
      [else '()])))

(define parse-variable
  (lambda (ls)
    (make-variable
     (car ls)                              ; number                 
     (parse-address (cadr ls))             ; next                   
     (caddr ls)			           ; name
     (list-ref ls 3)                       ; unmangled_name
     (list-ref ls 4)		           ; decl_position_seq      
     (list-ref ls 5)		           ; decl_position_column   
     (convert-bool (list-ref ls 6))        ; referenced             
     (convert-bool (list-ref ls 7))        ; needed                 
     (convert-bool (list-ref ls 8))        ; is_local_to_function   
     (list-ref ls 9)		           ; name_linkage
     (convert-bool (list-ref ls 10))        ; has_associated_pragma
     (convert-bool (list-ref ls 11))       ; static_used_by_instantiation
     (parse-address (list-ref ls 12))      ; type                   
     (parse-address (list-ref ls 13))      ; assoc_param_type       
     (list-ref ls 14)                      ; storage_class          
     (list-ref ls 15)		           ; declared_storage_class 
     (list-ref ls 16)		           ; init_kind              
     (convert-bool (list-ref ls 17))       ; address_taken          
     (parse-address (list-ref ls 18))      ; dynamic                
     (convert-bool (list-ref ls 19)) 	   ; param_value_has_been_changed   
     (convert-bool (list-ref ls 20))       ; param_used_more_than_once      
     (convert-bool (list-ref ls 21))       ; is_parameter           
     (convert-bool (list-ref ls 22))       ; is_handler_param
     (convert-bool (list-ref ls 23))       ; referenced_non_locally
     (convert-bool (list-ref ls 24))       ; is_partially_initialized
     (convert-bool (list-ref ls 25))       ; superseded_external
     (convert-bool (list-ref ls 26))       ; is_compound_literal
     (convert-bool (list-ref ls 27))       ; has_parenthesized_initializer
     (parse-address (list-ref ls 28))      ; declared_type
     (parse-address (list-ref ls 29))      ; constant      
     '()
     '()
     '())))

(define parse-field
  (lambda (ls)
    (make-field
     (car ls)                                 ; number                            
     (parse-address (cadr ls))     	      ;	next                              
     (caddr ls)				      ;	name                     
     (list-ref ls 3)			      ;	decl_position_seq                 
     (list-ref ls 4)			      ;	decl_position_column              
     (convert-bool (list-ref ls 5))	      ;	is_class_member      
     `()
     (convert-bool (list-ref ls 7))	      ;	referenced                        
     (convert-bool (list-ref ls 8))	      ;	needed                
     (list-ref ls 9)			      ;	name_linkage           
     (parse-address (list-ref ls 10))	      ;	type                              
     (list-ref ls 11)			      ;	offset
     (parse-address (list-ref ls 12)))))      ; assoc_type
     

(define parse-dynamic_init
  (lambda (ls)
    ;(printf "in pdi with ~s~n" ls)
    (make-dynamic_init
     (car ls)
     (parse-address (cadr ls))
     (caddr ls)
     (parse-address (list-ref ls 3))
     (parse-address (list-ref ls 4))
     )))

(define parse-param_type
  (lambda (ls)
    (make-param_type
     (car ls)                             ; number               
     (parse-address (cadr ls))            ; next                 
     (parse-address (caddr ls))           ; type
     'untyped)))                          ; dilated_type
;      (list-ref ls 3)                      ; name                 
;      (parse-address (list-ref ls 4))      ; default_arg_expr     
;      (list-ref ls 5)                   ; qualifiers

; (define parse-constant
;   (lambda (ls)
;     (make-constant
;      (car ls)                                   ; number                         
;      (parse-address (cadr ls))     	        ; next
;      (parse-address (caddr ls)) 		; type                           
;      (list-ref ls 3)				; kind
;      )))
;      (list-ref ls 4)                            ; length
;      (list-ref ls 5)			        ; value
;      (list-ref ls 6)                            ; integer_value
;      (convert-bool (list-ref ls 7))             ; implicit_cast
;      (list-ref ls 8)                            ; address_kind
;      (list-ref ls 9)                            ; address_offset
;      (parse-address (list-ref ls 10))           ; constant
;      (convert-bool (list-ref ls 11))	        ; referenced                      
;      (convert-bool (list-ref ls 12))	        ; needed                          
;      (convert-bool (list-ref ls 13))		; non_arithmetic                 
;      (convert-bool (list-ref ls 14))		; is_simple_zero
;      (parse-address (list-ref ls 15))           ; variable
;      (list-ref ls 16)                           ; float_value
;      (convert-bool (list-ref ls 17)))))	        ; null_pointer_constant_ruled_out 

(define parse-constant
  (lambda (ls)
  ;  (printf "Parsing constant list of length ~s~n" (length ls))
    (make-constant
     (car ls)                                                ; number                         
     (parse-address (cadr ls))				     ; next                           
     (caddr ls)						     ; name                           
     (list-ref ls 3)					     ; unmangled_name                 
     (list-ref ls 4)					     ; decl_position_seq              
     (list-ref ls 5)					     ; decl_position_column           
     (convert-bool (list-ref ls 6))			     ; referenced                     
     (convert-bool (list-ref ls 7))			     ; needed                         
     (convert-bool (list-ref ls 8))			     ; is_local_to_function           
     (list-ref ls 9)					     ; name_linkage                   
     (convert-bool (list-ref ls 10))			     ; has_associated_pragma          
     (convert-bool (list-ref ls 11))			     ; static_used_by_instantiation   
     (parse-address (list-ref ls 12))			     ; type                           
     (parse-address (list-ref ls 13))			     ; expr                           
     (convert-bool (list-ref ls 14))			     ; implicit_cast                  
     (convert-bool (list-ref ls 15))			     ; explicit_cast_applied          
     (convert-bool (list-ref ls 16))			     ; is_reinterpret_cast            
     (convert-bool (list-ref ls 17))			     ; non_arithmetic                 
     (convert-bool (list-ref ls 18))			     ; is_simple_zero                 
     (convert-bool (list-ref ls 19))			     ; null_pointer_constant_ruled_out
     (list-ref ls 20)					     ; kind                           
     (parse-constant_tail (list-ref ls 20) (list-tail ls 21))))) ; constant_tail

(define parse-constant_tail
  (lambda (kind ls)
    (case kind
      [(ck_integer) (make-integer_constant_tail
		     (car ls))] ; integer_value
      [(ck_string) (make-string_constant_tail
		    (car ls)       ; length
		    (cadr ls))]    ; value
      [(ck_float) (make-float_constant_tail
		   (car ls))] ; float_value
      [(ck_address) (make-address_constant_tail
		     (car ls)     ; address_kind
		     (cadr ls)    ; address_offset
		     (parse-address (caddr ls)))] ; address
      [(ck_aggregate) (make-aggregate_constant_tail
		       (parse-address (car ls))      ; first_constant
		       (parse-address (cadr ls)))]   ; last_constant
      [else '()])))

(define parse-expression
  (lambda (ls)
    (make-expression
     (car ls)                                               ; number                          
     (parse-address (cadr ls))			            ; next
     (caddr ls)					    ; kind                            
     (list-ref ls 3)                                ; operation_kind
     (parse-address (list-ref ls 4))   ; routine
     (parse-address (list-ref ls 5))			            ; type                           
     (parse-address (list-ref ls 6))	; operands
     (parse-address (list-ref ls 7))          ; variable
     (parse-address (list-ref ls 8))          ; constant
     (convert-bool (list-ref ls 9))			    ; result_is_not_used              
     (convert-bool (list-ref ls 10))		; compiler_generated
     (parse-address (list-ref ls 11))
     '() 
     '()
     '()
     '()
     #f
)))    ; field

(define parse-statement
  (lambda (ls)
    (make-statement
     (car ls)                           ; number               
     (parse-address (cadr ls))	        ; next                 
     (caddr ls)				; kind                 
     (list-ref ls 3)			; position
     (list-ref ls 4) ; final_position
     (parse-address (list-ref ls 5))      ; expr              
     (parse-address (list-ref ls 6 ))         ; statements
     (parse-address (list-ref ls 7))  ; assoc_scope              		     
     (parse-address (list-ref ls 8))	      ; lifetime
     (convert-bool (list-ref ls 9)) ; end_of_block_reachable   
     (parse-address (list-ref ls 10))     ; then_statement    
     (parse-address (list-ref ls 11))    ; else_statement
     (parse-address (list-ref ls 12))         ; initialization     
     (parse-address (list-ref ls 13))	    ; statement          
     (parse-address (list-ref ls 14))	    ; increment          
     (convert-bool (list-ref ls 15)) ; has_empty_else_clause
     (parse-address (list-ref ls 16))
     (parse-address (list-ref ls 17))
     #f)))    ; dynamic_init


(define parse-routine
  (lambda (ls)
    (make-routine
     (car ls)                                   ; number                        
     (parse-address (cadr ls))			; next                          
     (caddr ls)					; name                          
     (list-ref ls 3)				; unmangled_name                
     (list-ref ls 4)				; decl_position_seq             
     (list-ref ls 5)				; decl_position_column          
     (convert-bool (list-ref ls 6))		; referenced                    
     (convert-bool (list-ref ls 7))		; needed                        
     (convert-bool (list-ref ls 8))		; is_local_to_function          
     (list-ref ls 9)				; name_linkage                  
     (parse-address (list-ref ls 10))           ; type
     (list-ref ls 11)				; storage_class                 
     (convert-bool (list-ref ls 12))		; defined                       
     (convert-bool (list-ref ls 13))		; is_called                     
     (convert-bool (list-ref ls 14))		; definition_needed             
     (list-ref ls 15)				; assoc_scope                   
     (convert-bool (list-ref ls 16))		; compiler_generated            
     (convert-bool (list-ref ls 17))		; called                        
     '())))

(define parse-label
  (lambda (ls)
    ;(printf "parsing a label: ~s~n" ls)
    (make-label
     (car ls)                                   ; number                        
     (parse-address (cadr ls))			; next
     (caddr ls)                                 ; name              
     (list-ref ls 3)				; decl_position_seq             
     (list-ref ls 4)				; decl_position_column 
     (convert-bool (list-ref ls 5))		; referenced                    
     (convert-bool (list-ref ls 6))		; needed                 
     (convert-bool (list-ref ls 7))             ; is_local_to_function
     (list-ref ls 8)				; name_linkage   
     (convert-bool (list-ref ls 9))             ; reachable_by_fall_through
     (convert-bool (list-ref ls 10))            ; break_label
     (parse-address (list-ref ls 11)))))        ; exec_stmt

     

; if input is TRUE, return #t.  If input is FALSE or ~, return #f.
(define convert-bool
  (lambda (x)
    (case x
      [(TRUE) #t]
      [(FALSE ~) #f]
      [else (error 'convert-bool "~s is not an expected input." x)])))

; if ls is a list of length three, make an address from it.  If not, make an address from (0 0 0)
; (Of course, the default behavior can be changed as needed.)
(define parse-address
  (lambda (ls)
    (if (and (list? ls) (= 3 (length ls)))
	(apply make-address ls)
	(make-address 0 0 0))))
	     
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;; stuff to link up the records into a tree-like thing
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; assuming that the lists are sorted with contiguous indices, this is
; trivial (Just converting from one-based to zero-based.) However,
; since odd things could happen, I've put in error checking for
; now. If I need to I might make it actually search a bit later.
;ls is a list of records, i is the index we want, get-index is the
;accessor to get the index from the record
(define find-record
  (lambda (ls i get-index)
    (let ([rec (list-ref ls (- i 1))])
;      (if (= (get-index rec) i)
	  rec)))
	;  (begin (set! x ls)
	;  (error 'find-record "Desired index ~s does not match record: ~s~nin ~s" i rec ls))))))


(define build-tree-from-file
  (lambda (filename)
    (let ([sc-str (read-scil-list filename)])
      (let ([file-scope (scil_structure-file_scope sc-str)]
	    [fun-mem-list (scil_structure-fun_mem_region_list sc-str)])
	(link-il_header! (scil_structure-il_header sc-str) file-scope)
	(link-memory_region! file-scope file-scope)
	(map (lambda (x)
	       (link-memory_region! x file-scope))
	     fun-mem-list)
	sc-str
	))))

(define link-il_header!
  (lambda (ilh fs)
    (link-address! (il_header-primary_source_file ilh) fs)
    (link-address! (il_header-primary_scope ilh) fs)
    (link-address! (il_header-main_routine ilh) fs)
    (link-address! (il_header-nontag_types_used_in_exception_or_rtti ilh) fs)))

(define link-memory_region!
  (let ([link-mem!
	 (lambda (mr1 mr2)
	   (map (lambda (x)
		  (link-source_file! x mr2))
		(memory_region-source_file_list mr1))
	   (map (lambda (x)
		  (link-scope! x mr2))
		(memory_region-scope_list mr1))
	   (map (lambda (x)
		  (link-pragma! x mr2))
		(memory_region-pragma_list mr1))
	   (map (lambda (x)
		  (link-type! x mr2))
		(memory_region-type_list mr1))
	   (map (lambda (x)
		  (link-variable! x mr2))
		(memory_region-variable_list mr1))
	   (map (lambda (x)
		  (link-routine! x mr2))
		(memory_region-routine_list mr1))
	   (map (lambda (x)
		  (link-field! x mr2))
		(memory_region-field_list mr1))
	   (map (lambda (x)
		  (link-dynamic_init! x mr2))
		(memory_region-dynamic_init_list mr1))
	   (map (lambda (x)
		  (link-param_type! x mr2))
		(memory_region-param_type_list mr1))
	   (map (lambda (x)
		  (link-constant! x mr2))
		(memory_region-constant_list mr1))
	   (map (lambda (x)
		  (link-statement! x mr2))
		(memory_region-statement_list mr1))
	   (map (lambda (x)
		  (link-expression! x mr2))
		(memory_region-expression_list mr1))
	   (map (lambda (x)
		  (link-label! x mr2))
		(memory_region-label_list mr1)))])
    (lambda (mem-region file-scope)
      (if (not (equal? (memory_region-name mem-region)
		       (memory_region-name file-scope)))
	  (link-mem! mem-region mem-region))
      (link-mem! mem-region file-scope))))

(define link-source_file!
  (lambda (sf mr)
    (link-address! (source_file-next sf) mr)
    (link-address! (source_file-first_child_file sf) mr)
    (link-address! (source_file-last_child_file sf) mr)))

(define link-label!
  (lambda (l mr)
    (link-address! (label-next l) mr)
    (link-address! (label-exec_stmt l) mr)))

(define link-scope!
  (lambda (s mr)
    (link-address! (scope-next s) mr)
    (link-address! (scope-assoc_block s) mr)
    (link-address! (scope-lifetime s) mr)
    (link-address! (scope-constants s) mr)
    (link-address! (scope-types s) mr)
    (link-address! (scope-variables s) mr)
    (link-address! (scope-nonstatic_variables s) mr)
    (link-address! (scope-labels s) mr)
    (link-address! (scope-routines s) mr)
    (link-address! (scope-asm_entries s) mr)
    (link-address! (scope-scopes s) mr)
    (link-address! (scope-dynamic_init s) mr)
; not handling pragmas    (link-address! (scope-pragmas s) mr)
    (link-scope_tail! (scope-scope_tail s) mr)
    (link-address! (scope-labels s) mr)))

(define link-scope_tail!
  (lambda (st mr)
    (cond [(null? st) #t] ; nothing to link
	  [(func_prototype_scope_tail? st)
	   (link-address! (func_prototype_scope_tail-assoc_type st) mr)]
	  [(function_scope_tail? st)
	   (begin
	     (link-address! (function_scope_tail-routine_ptr st) mr)
	     (link-address! (function_scope_tail-parameters st) mr)
	     (link-address! (function_scope_tail-return_value_variable st) mr))]
	  [(class_struct_union_scope_tail? st)
	   (link-address! (class_struct_union_scope_tail-using_decls st) mr)]
	  [else 'link-scope_tail! "Unsupported scope tail: ~s" st])))

; does nothing for now
(define link-pragma!
  (lambda (p mr)
    #f))

(define link-type!
  (lambda (ty mr)
    (link-address! (type-next ty) mr)
    (link-address! (type-typeinfo_var ty) mr)
    (link-address! (type-reference ty) mr)
    (link-address! (type-pointer ty) mr)
    (link-address! (type-qualified ty) mr)
    (link-type_tail! (type-type_tail ty) mr)))

(define link-type_tail!
  (lambda (tt mr)
    (cond [(null? tt) #t] ; nothing to link
	  [(int_type_tail? tt)
	   (begin
	     (link-address! (int_type_tail-enum_info_constant_list tt) mr)
	     (link-address! (int_type_tail-enum_info_affiliated_type tt) mr))]
	  [(float_type_tail? tt) #t] ; nothing to link
	  [(pointer_type_tail? tt)
	   (link-address! (pointer_type_tail-type_pointed_to tt) mr)]
	  [(routine_type_tail? tt)
	   (begin
	     (link-address! (routine_type_tail-return_type tt) mr)
	     (link-address! (routine_type_tail-param_type_list tt) mr)
	     (link-address! (routine_type_tail-assoc_routine tt) mr)
	     (link-address! (routine_type_tail-prototype_scope tt) mr))]
	  [(array_type_tail? tt)
	   (begin
	     (link-address! (array_type_tail-element_type tt) mr)
	     (link-address! (array_type_tail-element_count_expr tt) mr)
	     (link-address! (array_type_tail-bound_constant tt) mr))]
	  [(struct_type_tail? tt)
	   (link-address! (struct_type_tail-field_list tt) mr)]
	  [(typeref_type_tail? tt)
	   (link-address! (typeref_type_tail-type_ref_type tt) mr)]
	  [(ptr_to_member_type_tail? tt)
	   (link-address! (ptr_to_member_type_tail-class_of_which_a_member tt) mr)]
	  [else (error 'link-type_tail! "Unsupported type tail: ~s" tt)])))

(define link-variable!
  (lambda (v mr)
    (link-address! (variable-next v) mr)
    (link-address! (variable-type v) mr)
    (link-address! (variable-assoc_param_type v) mr)
    (link-address! (variable-dynamic v) mr)
    (link-address! (variable-declared_type v) mr)))

(define link-routine!
  (lambda (r mr)
    (link-address! (routine-next r) mr)
    (link-address! (routine-type r) mr)))

(define link-field!
  (lambda (f mr)
    (link-address! (field-next f) mr)
    (link-address! (field-type f) mr)
    (link-address! (field-assoc_type f) mr)
    ;(link-address! (field-bit_size_constant f) mr)
  ))

(define link-dynamic_init!
  (lambda (di mr)
    (link-address! (dynamic_init-next di) mr)
    (link-address! (dynamic_init-variable di) mr)
    (link-address! (dynamic_init-constant di) mr)))

(define link-param_type!
  (lambda (pt mr)
    (link-address! (param_type-next pt) mr)
    (link-address! (param_type-type pt) mr)))
    

(define link-constant!
  (lambda (c mr)
    (link-address! (constant-next c) mr)
    (link-address! (constant-type c) mr)
    (link-address! (constant-expr c) mr)
    (link-constant_tail! (constant-constant_tail c) mr)))

(define link-constant_tail!
  (lambda (ct mr)
    (cond [(null? ct) #t] ; nothing to link
	  [(integer_constant_tail? ct) #t] ; nothing to link
	  [(string_constant_tail? ct) #t] ; nothing to link
	  [(float_constant_tail? ct) #t] ; nothing to link
	  [(address_constant_tail? ct)
	   (link-address! (address_constant_tail-address ct) mr)]
	  [(aggregate_constant_tail? ct)
	   (begin (link-address! (aggregate_constant_tail-first_constant ct) mr)
		  (link-address! (aggregate_constant_tail-last_constant ct) mr))]
	  [else (error 'link-constant_tail! "Unsupported constant tail")])))
	  

(define link-statement!
  (lambda (s mr)
    (link-address! (statement-next s) mr)
    (link-address! (statement-expr s) mr)
    (link-address! (statement-statements s) mr)
    (link-address! (statement-assoc_scope s) mr)
    (link-address! (statement-lifetime s) mr)
    (link-address! (statement-then_statement s) mr)
    (link-address! (statement-else_statement s) mr)
    (link-address! (statement-initialization s) mr)
    (link-address! (statement-statement s) mr)
    (link-address! (statement-increment s) mr)
    (link-address! (statement-dynamic_init s) mr)
    (link-address! (statement-label s) mr)))

(define link-expression!
  (lambda (e mr)
    (link-address! (expression-next e) mr)
    (link-address! (expression-routine e) mr)
    (link-address! (expression-type e) mr)
    (link-address! (expression-operands e) mr)
    (link-address! (expression-variable e) mr)
    (link-address! (expression-constant e) mr)
    (link-address! (expression-field e) mr)))


; this is what actually does the linking.
(define link-address!
  (let ([get-list-func
	 (lambda (a)
	   (case (address-field a)
	     [(source-file) memory_region-source_file_list]
	     [(scope) memory_region-scope_list]
	     [(pragma) memory_region-pragma_list]
	     [(type) memory_region-type_list]
	     [(variable) memory_region-variable_list]
	     [(routine) memory_region-routine_list]
	     [(field) memory_region-field_list]
	     [(dynamic-init) memory_region-dynamic_init_list]
	     [(param-type) memory_region-param_type_list]
	     [(constant) memory_region-constant_list]
	     [(statement) memory_region-statement_list]
	     [(expression expr-node) memory_region-expression_list]
	     [(label) memory_region-label_list]
	     [else (error 'link-address "address-field ~s not supported" (address-field a))]))]
	  [get-index-func
	   (lambda (a)
	     (case (address-field a)
	       [(source-file) source_file-number]
	       [(scope) scope-number]
	       [(pragma) pragma-number]
	       [(type) type-number]
	       [(variable) variable-number]
	       [(routine) routine-number]
	       [(field) field-number]
	       [(dynamic-init) dynamic_init-number]
	       [(param-type) param_type-number]
	       [(constant) constant-number]
	       [(statement) statement-number]
	       [(expression expr-node) expression-number]
	       [(label) label-number]
	       [else (error 'link-address "address-field ~s not supported" (address-field a))]))]
	  [name-okay?
	   (lambda (an mn)
	     (if (equal? an 'func-scope)
		 (not (equal? mn 'file-scope))
		 (equal? an mn)))])
    (lambda (a m)
      ;(printf "Link address ~s in mem region ~s:  ~n~n" a (memory_region-name m))
      (cond [(eqv? (address-mem_region_name a) 0)
	;     (printf "Address was null, didn't link~n")
	     #f] ; if we have a null address, don't link
	    [(not (null? (address-ref a)))
	 ;    (printf "Address already linked~n")
	     #f] ; if we're already linked, don't link
;	    [(equal? (address-mem_region_name a) (memory_region-name m))
	    [(name-okay? (address-mem_region_name a) (memory_region-name m))
	     (set-address-ref! a
			       (find-record ((get-list-func a) m) (address-num a) (get-index-func a)))
	  ;   (printf "linking complete~n")
	     #t]
;	    [else (error 'link-address! "Wasn't null, but didn't link")]))))
;	    [else (printf "Didn't link ~s in ~s~n" a (memory_region-name m))]))))
	    [else
	   ;  (printf "didn't link, unknown reason~n")
	     #f]))))


