;; -----------------------------------------------------------------------
 ; Indiana University Opie Lab Software License, Version 1.2
 ;
 ; Copyright (C) 2002 The Trustees of Indiana University.
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without
 ; modification, are permitted provided that the following conditions are
 ; met:
 ;
 ; 1) All redistributions of source code must retain the above
 ;    copyright notice, the list of authors in the original source
 ;    code, this list of conditions and the disclaimer listed in this
 ;    license;
 ;
 ; 2) All redistributions in binary form must reproduce the above
 ;    copyright notice, this list of conditions and the disclaimer
 ;    listed in this license in the documentation and/or other
 ;    materials provided with the distribution;
 ;
 ; 3) Any documentation included with all redistributions must include
 ;    the following acknowledgement:
 ;
 ;      "This product includes software developed by the Indiana
 ;      University Opie Lab.  For further information please visit
 ;      http://www.cs.indiana.edu/~dswise"
 ;
 ;    Alternatively, this acknowledgment may appear in the software
 ;    itself, and wherever such third-party acknowledgments normally
 ;    appear.
 ;
 ; 4) The name "Indiana University" or "Indiana University
 ;    Opie Lab" shall not be used to endorse or promote
 ;    products derived from this software without prior written
 ;    permission from Indiana University.  For written permission,
 ;    please contact http://www.cs.indiana.edu/~dswise.
 ;
 ; 5) Products derived from this software may not use "Indiana
 ;    Univeristy" name nor may "Indiana University" appear in their name,
 ;    without prior written permission of the Indiana University.
 ;
 ; Indiana University provides no reassurances that the source code
 ; provided does not infringe the patent or any other intellectual
 ; property rights of any other entity.  Indiana University disclaims any
 ; liability to any recipient for claims brought by any other entity
 ; based on infringement of intellectual property rights or otherwise.
 ;
 ; LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH
 ; NO WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA
 ; UNIVERSITY GIVES NO WARRANTIES AND MAKES NO REPRESENTATION THAT
 ; SOFTWARE IS FREE OF INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR
 ; OTHER PROPRIETARY RIGHTS.  INDIANA UNIVERSITY MAKES NO WARRANTIES THAT
 ; SOFTWARE IS FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP
 ; DOORS", "WORMS", OR OTHER HARMFUL CODE.  LICENSEE ASSUMES THE ENTIRE
 ; RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR ASSOCIATED MATERIALS,
 ; AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION GENERATED USING
 ; SOFTWARE.
;; ------------------------------------------------------------------------

; scil-to-c.ss
; created by Brooke Chenoweth

; take the scil tree and generate c code

; I do this by building up a nested list of strings while traversing
; the scil_structure and then flattening the list and string-appending
; the strings together at the end.

; scil-parse.ss (and thus, scil-records.ss)  must be loaded before this file

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; (generate-c-from-scil infilename) returns the string of C code corresponding to the scil stuff in file infilename
; (generate-c-from-scil infilename outfilename) writes the C code to file outfilename
(define generate-c-from-scil
  (lambda (infilename . out)
    (let ([handled-sstr (handle-scil_structure (build-tree-from-file infilename))])
      (if (null? out)
	  handled-sstr
	  (with-output-to-file (car out)
	    (lambda ()
	      handled-sstr))))))

; this isn't a very pretty way of handling things, but I didn't
; discover that the routine assoc_scope field was just a memory region
; number until I'd already handled creating the records and building
; the tree.  It might be a good idea to change things later so we don't
; need this global variable.
(define *mem-region-list* '())

(define get-mem-region
  (lambda (n)
    (let loop ([ls *mem-region-list*])
      (cond [(null? ls) #f] ;(error 'get-mem-region "Memory region ~s not found" n)]
	    [(= n (memory_region-number (car ls))) (car ls)]
	    [else (loop (cdr ls))]))))


; Take a scil_structure and return the string of C code generated from it
(define handle-scil_structure
  (lambda (sstr)
    (set! *mem-region-list* (cons (scil_structure-file_scope sstr) (scil_structure-fun_mem_region_list sstr)))
    (let ([result (flatten-and-string-append (handle-il_header (scil_structure-il_header sstr)))])
      (set! *mem-region-list* '())
      result)))

; flatten-and-string-append takes any number of arguments, assumed to
; be strings or lists (possibly deep) of strings and concatenates the
; strings together in the order they appear
(define flatten-and-string-append
  (letrec ([flatstr (lambda (x)
		      ;(printf "entered flatstr with: ~s~n" x)
		      (cond [(string? x) x]
			    [(null? x) ""]
			    [(pair? x) (string-append (flatstr (car x))
						      (flatstr (cdr x)))]
			    [(number? x) (number->string x)]
			    [(symbol? x) (symbol->string x)]
			    [else (error 'flatten-and-string-append "~s isn't a list or string." x)]))])
    (lambda ls
      (flatstr ls))))

(define handle-il_header
  (lambda (ilh)
    (handle-scope-ref (il_header-primary_scope ilh))))

(define handle-scope-ref
  (lambda (sr)
    (if (null? (address-ref sr))
	'() ;"/* null scope ref */"
	(handle-scope (address-ref sr)))))

; returns first non-pair element in a nested list
(define first-elem
  (lambda (ls)
    (if (pair? (car ls))
	(first-elem (car ls))
	(car ls))))

; returns a nested list with first non-pair element removed
(define other-elems
  (lambda (ls)
    (if (pair? (car ls))
	(cons (other-elems (car ls)) (cdr ls))
	(cdr ls))))
     
(define handle-typedef
  (lambda (t)
    ;(printf "handle-typedef with ~s~n" t)
    (if (equal? (type-kind t) 'tk_typeref)
	`("typedef " 
	  ,(handle-type (a-r (typeref_type_tail-type_ref_type (type-type_tail t))))
	  " "
	  ,(type-name t)
	  ";\n")
	"")))

(define handle-scope
  (lambda (s)
    (let ([k (scope-kind s)]
	  [var-decls (map (lambda (v)
			    (let ([vstr (handle-type-ref (variable-type v))]
				  [sc (handle-storage-class
				       (variable-storage_class v))])
;*******************************************************************************************
;CHANGE MADE HERE
;***********************			      
			      (if (not (equal? (variable-init_kind v) "initk_dynamic"))
				  (if (eqv? 'tk_array (type-kind (address-ref (variable-type v)))) ; HACK!
				      (list (first-elem vstr) " " (symbol->string (variable-name v))
					    (other-elems vstr) ";" (string #\newline))
				      (list  sc
					     vstr " "
					     (symbol->string (variable-name v))
					     ";" (string #\newline)))
				  "")))
;********************************************************************************************
			    (get-list-of-records (scope-variables s) variable-next))]
	  [nonstatic-var-decls 
; 	   (map (lambda (v)
; 		  (list (handle-type-ref (variable-type v)) " "
; 			(symbol->string (variable-name v))
; 			";" (string #\newline)))
; 		(get-list-of-records (scope-nonstatic_variables s) variable-next))
;******************************************************************************************
;CHANGE MADE HERE
;*******************************
	   (map (lambda (v) 
		  (if (not (equal? (variable-init_kind v) "initk_dynamic"))
		      (list (handle-type-ref/var (variable-type v) v) ";"
			    (string #\newline)) 
		      ""))
;*******************************************************************************************
		(get-list-of-records (scope-nonstatic_variables s) variable-next))]
	  [routine-list (get-list-of-records (scope-routines s)
					     routine-next)])
      (case k
	[(sck_file) (list var-decls
			  (map handle-typedef (get-list-of-records (scope-types s) type-next))
			  (map handle-routine-decl routine-list)
			  (map (lambda (r)
				 (if (routine-defined r) (handle-routine-def r) '()))
			       routine-list))]
	[(sck_block) (list "{" (string #\newline)
			   var-decls
			   nonstatic-var-decls
			   (handle-statement-ref (scope-assoc_block s)) (string #\newline)
			   "}" (string #\newline))]
	[(sck_func_prototype) "/* sck_func_prototype */"]
	[(sck_class_struct_union) "/* sck_class_struct_union */"]
	[(sck_condition) "/* sck_condition */"]
	[(sck_namespace) "/* sck_namespace */"]
	[(sck_function) (list "{" (string #\newline)
			   var-decls
			   nonstatic-var-decls
			   (handle-statement-ref (scope-assoc_block s)) (string #\newline)
			   "}" (string #\newline))]
	[(sck_template_declaration) "/* sck_template_declaration */"]
	[(sck_template_instantiation) "/* sck_template_instantiation */"]
	[else (error 'handle-scope "Unsupported scope kind ~s" k)]))))

; takes a record address and accessor for "next" field and returns the list
; of records found by following "next" until the address is null
(define get-list-of-records
  (lambda (addr get-next)
    (let ([r (a-r addr)])
      (if (null? r)
	  '()
	  [routine-list (get-list-of-records (scope-routines s) routine-next)])
      (case k
	[(sck_file) (list var-decls
			  (map handle-routine-decl routine-list)
			  (map (lambda (r)
				 (if (routine-defined r) (handle-routine-def r) '()))
			       routine-list))]
	[(sck_block) (list "{" (string #\newline)
			   var-decls
			   nonstatic-var-decls
			   (handle-statement-ref (scope-assoc_block s)) (string #\newline)
			   "}" (string #\newline))]
	[(sck_func_prototype) "/* sck_func_prototype */"]
	[(sck_class_struct_union) "/* sck_class_struct_union */"]
	[(sck_condition) "/* sck_condition */"]
	[(sck_namespace) "/* sck_namespace */"]
	[(sck_function) (list "{" (string #\newline)
			   var-decls
			   nonstatic-var-decls
			   (handle-statement-ref (scope-assoc_block s)) (string #\newline)
			   "}" (string #\newline))]
	[(sck_template_declaration) "/* sck_template_declaration */"]
	[(sck_template_instantiation) "/* sck_template_instantiation */"]
	[else (error 'handle-scope "Unsupported scope kind ~s" k)]))))

; takes a record address and accessor for "next" field and returns the list
; of records found by following "next" until the address is null
(define get-list-of-records
  (lambda (addr get-next)
    (let ([r (a-r addr)])
      (if (null? r)
	  '()
	  (cons r (get-list-of-records (get-next r) get-next))))))

(define handle-type-ref
  (lambda (tr)
    (if (null? (address-ref tr))
	"/* null type ref */"
	(handle-type (address-ref tr)))))

(define handle-type
  (lambda (t)
    (case (type-kind t)
      [(tk_error) "/* tk_error */"]
      [(tk_unknown) "/* tk_unknown */"]
      [(tk_void) "void"]
      [(tk_integer) (handle-int (type-type_tail t))]
      [(tk_float) (handle-float (type-type_tail t))]
      [(tk_pointer) (let ([ptt (type-type_tail t)])
		      (list (handle-type-ref (pointer_type_tail-type_pointed_to ptt)) "* "))]
      [(tk_routine) (let ([rtt (type-type_tail t)])
		      (cons (handle-type-ref (routine_type_tail-return_type rtt))
			    (map handle-param_type
				 (get-list-of-records (routine_type_tail-param_type_list rtt) param_type-next))))]
      [(tk_array)  (let* ([att (type-type_tail t)])
		     (list (handle-type-ref (array_type_tail-element_type att))
			   "[" (number->string (array_type_tail-number_of_elements att)) "]"))]
      [(tk_class) "/* tk_class */"]
      ;[(tk_struct) "/* tk_struct */"]
;****************************************************************************************
; CHANGE MADE HERE
;****************************************************************************************
      [(tk_struct) ;;"/* tk_struct */"]
       (let* ([stt (type-type_tail t)]
	      [fldlst (struct_type_tail-field_list stt)])
	 (list "struct " "/* "(type-name t) " */" 
	       "{" (string #\newline)
	       (handle-struct fldlst) (string #\newline) "} " var))]
;****************************************************************************************
      [(tk_union) "/* tk_union */"]
      [(tk_typeref) (type-name t)]
      [(tk_ptr_to_member) "/* tk_ptr_to_member */"]
      [(tk_template_param) "/* tk_template_param */"]
      [else (error 'handle-type "unsupported type kind")])))

(define handle-type-ref/var
  (lambda (tr var)
    (if (null? (address-ref tr))
	"/* null type ref */"
	(handle-type/var (address-ref tr) var))))

(define handle-storage-class
  (lambda (sc)
    ;(printf "found storage class: ~s~n" sc)
    (case sc
      [(sc_auto) ""]
      [(sc_unspecified) ""]
      [(sc_extern) "extern "]
      [(sc_static) "static "]
      [(sc_register) "register "]
      [(sc_restrict) "restrict "]
      [else (error 'handle-storage-class "unknown storage class ~s" sc)])))

(define handle-type/var
  (lambda (t v)
    ;(printf "in handle-type/var with: ~s~n" v)
    (let ([var (handle-variable v)]
	  [sc (handle-storage-class (variable-storage_class v))])
      (case (type-kind t)
	[(tk_error) "/* tk_error */"]
	[(tk_unknown) "/* tk_unknown */"]
	[(tk_void) (list sc "void" var)]
;***********************************************************************************
;CHANGE MADE HERE
;*******************
	[(tk_integer) ;(if (not (equal? (variable-init_kind v) "initk_dynamic")) 
			  (list  sc (handle-int (type-type_tail t)) " " var)]
			  ;(handle-dynamic-init (a-r (variable-dynamic v))))]
			  ;"")]
	[(tk_float) ;(if (not (equal? (variable-init_kind v) "initk_dynamic")) 
			(list (handle-float (type-type_tail t)) " " var)]
			;"")]
;***********************************************************************************
	[(tk_pointer) (let ([ptt (type-type_tail t)])
			;(printf "calling handle-type-ref/var with: ~s~n" var)
			(list (handle-type-ref/var (pointer_type_tail-type_pointed_to ptt) v) "* "))]
	[(tk_routine) (let ([rtt (type-type_tail t)])
			(cons (handle-type-ref/var (routine_type_tail-return_type rtt) v)
			      (map handle-param_type
				   (get-list-of-records (routine_type_tail-param_type_list rtt) param_type-next))))]
;***********************************************************************************
; CHANGE MADE HERE: see scil-utilities.ss for definition of add-as-first-dimension
;***********************************************************************************
	[(tk_array)
	 (let* ([att (type-type_tail t)]
		[lst (list "[" (number->string
				(array_type_tail-number_of_elements
				 att)) "]")]
		[htrv (handle-type-ref/var (array_type_tail-element_type att) v)])
	   	       (add-as-first-dimension lst htrv))]
;************************************************************************************

	[(tk_class) "/* tk_class */"]

;************************************************************************************
;CHANGE MADE HERE
;*********************
	[(tk_struct) ;;"/* tk_struct */"]
	 (let* ([stt (type-type_tail t)]
		[fldlst (struct_type_tail-field_list stt)])
		(list "struct " "/* "(type-name t) " */" 
		      "{" (string #\newline)
		      (handle-struct fldlst) (string #\newline) "} " var))]
;*************************************************************************************
	[(tk_union) "/* tk_union */"]
	[(tk_typeref) (list (type-name t) " " var)]
	[(tk_ptr_to_member) "/* tk_ptr_to_member */"]
	[(tk_template_param) "/* tk_template_param */"]
	[else (error 'handle-type "unsupported type kind")]))))

(define handle-struct
  (lambda (fldlst-ref)
    (let* ([fldtp (handle-type-ref (field-type (a-r fldlst-ref)))]
	   [fldnm (field-name (a-r fldlst-ref))]
           [hasnext? (not (nil? (field-next (a-r fldlst-ref))))])
      (if hasnext?
	  (list fldtp " " fldnm ";" (string #\newline)  
		(handle-struct (field-next (a-r fldlst-ref))))
	  (list fldtp " " fldnm ";")))))

(define handle-int
  (lambda (i)
    (let ([k (int_type_tail-int_kind i)]) 
      (case k
	[(ik_signed_char) "signed char "]
	[(ik_unsigned_char) "unsigned char "]
	[(ik_short) "short "]
	[(ik_unsigned_short) "unsigned short "]
	[(ik_int) "int "]
	[(ik_unsigned_int) "unsigned int "]
	[(ik_long) "long "]
	[(ik_unsigned_long) "unsigned long "]
	[else k]))))
;	[else (error 'handle-int "Unknown int kind ~s" k)]))))

(define handle-float
  (lambda (f)
    (let ([k (float_type_tail-float_kind f)])
      (case k
	[(fk_float) "float "]
	[(fk_double) "double "]
	[else k]))))				
;	[else (error 'handle-float "Unknown float kind")]))))

(define handle-param_type
  (lambda (pt)
    (handle-type-ref (param_type-type pt))))


(define get-routine-name
  (lambda (r)
    (let ([unmangled (routine-unmangled_name r)]
	  [name (routine-name r)])
      (if (or (equal? unmangled '~)
	      (and (list? unmangled) (null? r)))
	  (symbol->string name)
	  (symbol->string unmangled)))))

; routines are a bit hacked right now.

(define map-two
  (lambda (f a b)
    (cond
     [(null? a) '()]
     [(null? b) '()]
     [(cons (f (car a) (car b)) (map-two f (cdr a) (cdr b)))])))

(define get-list-of-vars-with-shadows
  (lambda (vars)
    (let ([cur (a-r vars)])
      (if (nil? cur) 
	  '()
	  (let ([even  (a-r (variable-even_shadow cur))]
		[odd   (a-r (variable-odd_shadow cur))])
	    (printf "in glovws, even -- ~s~n               odd ~s~n" even odd)
	    (cond
	     [(not (nil? even))
	      (cons cur (cons even (get-list-of-vars-with-shadows (variable-next cur))))]
	     [(not (nil? odd))
	      (cons cur (cons odd (get-list-of-vars-with-shadows (variable-next cur))))]
	     [(cons cur (get-list-of-vars-with-shadows (variable-next cur)))]))))))

(define handle-routine-def
  (lambda (r)
    ;(printf "handle-routine~n")
    (let ([type (handle-type-ref (routine-type r))]
	  [name (get-routine-name r)]
	  [region_num (routine-assoc_scope r)])
      ; this is taking a lot for granted to make it work.
      ; this assumes that mr has one scope, which is a function_scope
      (let ([mr (get-mem-region region_num)])
	(let ([params (map handle-variable
			   (get-list-of-records
			    (function_scope_tail-parameters (scope-scope_tail  (car (memory_region-scope_list mr)))) variable-next))])
	  (set! param-temp params)
	  (set! type-temp type)
	  ;(printf "params: ~s~n" params)
	  (list (car type) " " name "("
		(if (null? (cdr type))
		    '()
		    (cons (list (cadr type) (car params))
			  (map-two (lambda (t v)
				 (list ", " t " " v))
			       (cddr type) (cdr params))))
		")" (string #\newline)
		(handle-memory_region mr) (string #\newline)))))))
  
(define delimit
  (lambda (xs delimiter)
    (cond 
      [(null? xs) '()]
      [(null? (cdr xs)) xs]
      [else (append `(,(car xs) ,delimiter) (delimit (cdr xs) delimiter))])))

(define handle-routine-decl
  (lambda (r)
    (let ([type (handle-type-ref (routine-type r))]
	  [name (get-routine-name r)]
	  [region_num (routine-assoc_scope r)])
; this is taking a lot for granted to make it work.
; this assumes that mr has one scope, which is a function_scope
      (cond
       [(equal? name "Opie415MatrixPragmaRoutine")
	"/* removed prototype for pragma routine generated by intermediate compile step */\n"]
       [(routine-compiler_generated r)
	`("/* did not generate function prototype for compiler generated routine "
	,name
	"*/\n")]
       [else
	(let ([mr (get-mem-region region_num)])
	  (if mr
	      (let ([params (map handle-variable
				 (get-list-of-records
				  (function_scope_tail-parameters (scope-scope_tail  (car (memory_region-scope_list mr)))) variable-next))])
		(list (car type) " " name "("
		      (if (null? (cdr type))
			  '()
			    (cons (list (cadr type) (car params))
				  (map-two (lambda (t v)
					 (list ", " t " " v))
				       (cddr type) (cdr params))))
		      ");" (string #\newline)))
					;(string #\newline)
	      (let* ([type (a-r (routine-type r))]
		       [return-type (a-r (routine_type_tail-return_type (type-type_tail type)))]
		       [param-type (routine_type_tail-param_type_list (type-type_tail type))])
					;(printf "params ~s" param-type)
		(append 
		 `(,(if (equal? (routine-storage_class r) 'sc_extern) "extern " ""))
		 `(,(handle-type return-type) " ")
		 `(,(routine-unmangled_name r) "(")
		   (delimit
		    (map handle-type
			 (map a-r
			      (map 
			       param_type-type 
			       (get-list-of-records
				param-type
				param_type-next))))
		    ", ")
		   `(");\n")))))]))))
			

; this just handles all the scopes in the memory region's scope list
; this might not do what you want if the memory region has multiple scopes
(define handle-memory_region
  (lambda (mr)  ; mr is a memory_region or #f (if we didn't find the mem region we wanted)
    (if mr
	(map handle-scope (memory_region-scope_list mr))
	'())))

(define handle-dynamic-init
  (lambda (di)
    (let ([kind (dynamic_init-kind di)]
	  [var (a-r (dynamic_init-variable di))]
	  [const (a-r (dynamic_init-constant di))])
      ;(printf "~n~nkind: ~s~nvar: ~s~nconst: ~s~n~n" kind var const)
      (cond
       [(equal? kind `dik_constant)
	(cond
	 [(null? var)
	  (error 'handle-dynamic-init "dynamic-init of kind dik_constant has a null variable~n")]
	 [(null? const)
	  (error 'handle-dynamic-init "dynamic-init of kind dik_constant has a null constant~n")]
	 
;**********************************************************************************
;CHANGE MADE HERE 
;**********************************************************************************
	 [else
	  ;(printf "~n~n~s~n~n" const)
	  (if (equal? (type-kind (a-r (variable-type var))) 'tk_array)
	      (if (isastring? (handle-constant const))
		  `(,(handle-storage-class (variable-storage_class var)) 
		    ,(get-type-before-brackets (flatten (handle-type-ref (variable-type var)))) 
		    " " ,(handle-variable var) 
		    "[] = " ,(handle-constant const) ";\n")
		  `(,(handle-storage-class (variable-storage_class var)) 
		    ,(get-type-before-brackets (flatten (handle-type-ref (variable-type var)))) 
		    " " ,(handle-variable var)
		    ,(get-dimension (flatten (handle-type-ref (variable-type var)))) 
		    " = " ,(handle-constant const) ";\n"))
	      `(,(handle-storage-class (variable-storage_class var)) 
		,(handle-type-ref (variable-type var)) " " 
		,(handle-variable var) " = " ,(handle-constant const) ";\n"))])]
;************************************************************************************
       [(equal? kind 'dik_expression)
	"/* dyn_init of kind dik_expression */ \n" ]
	;`(,(handle-type-ref (variable-type var)) " " ,(handle-variable var) " = " )]
       [else
	(error 'handle-dynamic-init "unrecognized dynamic-init kind ~s~n" kind)]))))
	  

;*******************************************************************************
; BIG CHANGES MADE HERE
;------------------------------------------------------------------------------
; These changes are all to make aggregates work correctly.
;*******************************************************************************
(define flatten 
  (lambda (slist)
    ;(printf "~n~nentered flatten with ~s~n~n" slist)
    (cond 
     [(null? slist) '()]
     [(pair? (car slist)) (append (flatten (car slist)) (flatten (cdr slist)))]
     [(null? (car slist)) (flatten (cdr slist))]
     [else (append (list (car slist)) (flatten (cdr slist)))])))

(define get-dimension
  (lambda (ls)
    ;(printf "~n~nentered get-dimension with ~s~n~n" ls)
    (cond
     [(null? ls) "[]"]
     [(equal? (car ls) "[") (flip-dimensions ls)]
     [else (get-dimension (cdr ls))])))

(define flip-dimensions
  (lambda (ls)
    ;(printf "~n~nentered flip-dimension with ~s~n~n" ls)
    (letrec ([get-numbers
	      (lambda (list accum)
		(cond
		 [(null? list) accum]
		 [(or (equal? (car list) "[")
		      (equal? (car list) "]"))
		  (get-numbers (cdr list) accum)]
		 [else (get-numbers (cdr list) (cons (car list) accum))]))]
	     [add-brackets
	      (lambda (list accum)
		(cond
		 [(null? list) accum]
		 [else (add-brackets (cdr list) 
				     (append `("[" ,(car list) "]")
					     accum))]))])
      (add-brackets (reverse (get-numbers ls '())) '()))))

(define isastring?
  (lambda (x)
    (cond
     [(not (list? x)) (error 'isastring? "not a list")]
     [(null? x) (error 'isastring? "null list")]
     [(equal? (car x) "\"") #t]
     [else #f])))

;************************************************************************

(define handle-statement-ref
  (lambda (sr)
    (if (null? (address-ref sr))
	"" ;"/* null statement ref */"
	(handle-statement (address-ref sr)))))

(define handle-statement
  (lambda (s)
    (let ([kind (statement-kind s)]
	  [next (statement-next s)])
      (case kind
	[(stmk_empty) (list "{}" (string #\newline) (handle-statement-ref next))]
	[(stmk_expr) (list (handle-expression-ref (statement-expr s)) ";"
			   (if (null? (address-ref next)) ""
			       (list  (string #\newline) (handle-statement-ref next))))]
	[(stmk_return) (list "return " (handle-expression-ref (statement-expr s)) ";"
			     (string #\newline) (handle-statement-ref next))]
	[(stmk_if) (list "if( " (handle-expression-ref (statement-expr s)) " )" (string #\newline)
			 "{" (handle-statement-ref (statement-then_statement s)) "}" (string #\newline)
			 (if (statement-has_empty_else_clause s)
			     '()
			     (list "else {" (handle-statement-ref (statement-else_statement s)) "}"))
			 (string #\newline) (handle-statement-ref next))]
	[(stmk_while) (list "while( " (handle-expression-ref (statement-expr s)) ")" (string #\newline)
			    "{" (handle-statement-ref (statement-statement s)) "}" 
			    (string #\newline) (handle-statement-ref next))]
	[(stmk_end_test_while) "/* stmk_end_test_while */"]
	[(stmk_goto) (let* ([lab (a-r (statement-label s))]
			    [break? (label-break_label lab)])
			(list (if break?
				  (list "break;\n")
				  (list "goto " (label-name lab) ";\n"))
			      (handle-statement-ref next)))]
	[(stmk_label) (let* ([lab (a-r (statement-label s))]
			     [break? (label-break_label lab)])
			(list (if break?
				  (list "")
				  (list (label-name lab) ":\n"))
			      (handle-statement-ref next)))]
	[(stmk_block) (list  (handle-statement-ref (statement-statements s) ) " "
			     (string #\newline) (handle-statement-ref next) )]
	[(stmk_for) (list "for( " (handle-statement-ref (statement-initialization s)) 
			  (if (null? (a-r (statement-initialization s)))
			      "; " 
			      " ")
			  (handle-expression-ref (statement-expr s)) "; "
			  (handle-expression-ref (statement-increment s)) ")" (string #\newline)
			  "{ " (handle-statement-ref (statement-statement s)) " }"
			  (string #\newline) (handle-statement-ref next))]
	[(stmk_switch) "/* stmk_switch */"] 
	[(stmk_init) `(,(handle-dynamic-init (a-r (statement-dynamic_init s)))
		       ,(handle-statement-ref next))]
	[(stmk_asm) "/* stmk_asm */"]
	[(stmk_asm_func_body) "/* stmk_asm_func_body */"]
	[(stmk_try_block) "/* stmk_try_block */"]
	[(stmk_microsoft_try) "/* stmk_microsoft_try */"]
	[(stmk_decl) "/* stmk_decl */"]
	[(stmk_set_vla_size) "/* stmk_set_vla_size */"]
	[(stmk_vla_decl) "/* stmk_vla_decl */"]
	[(stmk_vla_dealloc) "/* stmk_vla_dealloc */"]
	; the following is used by the Opie compiler only
        ; i.e it will not be generated by EDG
        [(stmk_mybreak) (list "break;" (string #\newline))]
	[else (error 'handle-statement "Unsupported statement kind ~s" kind)]))))

(define handle-routine-addr
  (lambda (ra)
    (let ([r (address-ref ra)])
      (if (null? r)
	  ""
	  (get-routine-name r)))))

(define handle-expression-ref
  (lambda (er)
    (if (null? (address-ref er))
	'() ;"/* null expression ref */"
	(handle-expression (address-ref er)))))

(define handle-expression
  (lambda (exp)
    (let ([k (expression-kind exp)]
	  [next (expression-next exp)])
      (case k
	[(enk_error) "/* enk_error */"]
	[(enk_operation) (handle-operation exp)]
	[(enk_constant) (handle-constant-ref (expression-constant exp))]
	[(enk_variable) (handle-variable-ref (expression-variable exp))]
	[(enk_variable_address) `("&" ,(handle-variable-ref (expression-variable exp)))] ; this probably should be different
	[(enk_char_variable_length) "/* enk_char_variable_length */"]
	[(enk_routine_address) (handle-routine-addr (expression-routine exp))]
	[(enk_field) ;;"/* enk_field */"]
	 (handle-field-ref (expression-field exp))]
	[(enk_temp_init) "/* enk_temp_init */"]
	[(enk_new_delete) "/* enk_new_delete */"]
	[(enk_throw) "/* enk_throw */"]
	[(enk_condition) "/* enk_condition */"]
	[(enk_object_lifetime) "/* enk_object_lifetime */"]
	[(enk_typeid) "/* enk_typeid */"]
	[(enk_runtime_sizeof) "/* enk_runtime_sizeof */"]
	[(enk_address_of_ellipsis) "/* enk_address_of_ellipsis */"]
	[(enk_lowered_eh_constant) "/* enk_lowered_eh_constant */"]
	[(enk_result_of_overriding_function) "/* enk_result_of_overriding_function */"]
	[else (error 'handle-expression "Unsupported expression kind ~s" (expression-kind exp))]))))

(define handle-field-ref
  (lambda (fref)
    (field-name (a-r fref))))

(define handle-lvalue
  (lambda (exp)
    (let ([k (expression-kind exp)])
      (if (or (equal? k 'enk_variable)
	      (equal? k 'enk_variable_address))
	  (handle-variable-ref (expression-variable exp))
	  (handle-expression exp)))))

(define handle-operation
  (lambda (op-exp)
    (let ([k (expression-operation_kind op-exp)]
	  [ops (map handle-expression (get-list-of-records (expression-operands op-exp) expression-next))])
      ;(printf "entered handle-operation with k: ~s~n" k)
      ;(printf "operands: ~s~n" (expression-operands op-exp))
      ;(printf "handle-operation with ops: ~s~n" ops)
      (case k
	[(eok_indirect) (list "( *" (car ops) ")")]
	[(eok_inegate eok_fnegate) (list " -(" (car ops) ")")]
	[(eok_unary_plus) "/* eok_unary_plus */"]
	[(eok_not) (list "!(" (car ops) ")")]
	[(eok_cast) (list "((" (handle-type-ref (expression-type op-exp)) ")" (car ops) ")")]
	[(eok_base_class_cast) "/* eok_base_class_cast */"]
	[(eok_derived_class_cast) "/* eok_derived_class_cast */"]
	[(eok_pm_base_class_cast) "/* eok_pm_base_class_cast */"]
	[(eok_pm_derived_class_cast) "/* eok_pm_derived_class_cast */"]
	[(eok_lvalue_cast) "/* eok_lvalue_cast */"]
	[(eok_dynamic_cast) "/* eok_dynamic_cast */"]
	[(eok_bool_cast) "/* eok_bool_cast */"]
	[(eok_complement) "/* eok_complement */"]
	[(eok_ipost_incr eok_fpost_incr eok_ppost_incr) 
	 (list (handle-lvalue (a-r (expression-operands op-exp))) "++ ")]
	[(eok_ipost_decr eok_fpost_decr eok_ppost_decr) 
	 (list (handle-lvalue (a-r (expression-operands op-exp))) "-- ")]
	[(eok_ipre_incr eok_fpre_incr eok_ppre_incr) 
	 (list " ++" (handle-lvalue (a-r (expression-operands op-exp))))]
	[(eok_ipre_decr eok_fpre_decr eok_ppre_decr) 
	 (list " --" (handle-lvalue (a-r (expression-operands op-exp))))]
	[(eok_lvalue_from_struct_rvalue) "/* eok_lvalue_from_struct_rvalue */"]
	[(eok_assume) "/* eok_assume */"]
	[(eok_xnegate) "/* eok_xnegate */"]
	[(eok_char_length) "/* eok_char_length */"]
	[(eok_address_of_value) "/* eok_address_of_value */"]
	[(eok_loc) "/* eok_loc */"]
	[(eok_test_logical) "/* eok_test_logical */"]
	[(eok_iadd eok_fadd eok_padd) (list "(" (car ops) " + " (cadr ops) ")")]
	[(eok_isubtract eok_fsubtract eok_psubtract) (list (car ops) " - " (cadr ops)) ]
	[(eok_imultiply eok_fmultiply) (list "(" (car ops) " * " (cadr ops) ")")]
	[(eok_idivide eok_fdivide) (list (car ops) " / " (cadr ops))]
	[(eok_ieq eok_feq eok_peq) (list (car ops) " == " (cadr ops))]
	[(eok_ine eok_fne eok_pne) (list (car ops) " != " (cadr ops))]
	[(eok_igt eok_fgt eok_pgt) (list (car ops) " > " (cadr ops))]
	[(eok_ilt eok_flt eok_plt) (list (car ops) " < " (cadr ops))]
	[(eok_ige eok_fge eok_pge) (list (car ops) " >= " (cadr ops))]
	[(eok_ile eok_fle eok_ple) (list (car ops) " <= " (cadr ops))]
	[(eok_iassign eok_fassign eok_passign) 
	 (list (handle-lvalue (a-r (expression-operands op-exp))) " = " (cadr ops))]
	[(eok_xadd) "/* eok_xadd */"]
	[(eok_xsubtract) "/* eok_xsubtract */"]
	[(eok_xmultiply) "/* eok_xmultiply */"]
	[(eok_xdivide) "/* eok_xdivide */"]
	[(eok_xeq) "/* eok_xeq */"]
	[(eok_xne) "/* eok_xne */"]
	[(eok_xassign) "/* eok_xassign */"]
	[(eok_complex) "/* eok_complex */"]
	[(eok_ceq) "/* eok_ceq */"]
	[(eok_cne) "/* eok_cne */"]
	[(eok_cgt) "/* eok_cgt */"]
	[(eok_clt) "/* eok_clt */"]
	[(eok_cge) "/* eok_cge */"]
	[(eok_cle) "/* eok_cle */"]
	[(eok_cassign) "/* eok_cassign */"]
	[(eok_concat) "/* eok_concat */"]
	[(eok_i_to_i_expon) "/* eok_i_to_i_expon */"]
	[(eok_f_to_i_expon) "/* eok_f_to_i_expon */"]
	[(eok_x_to_i_expon) "/* eok_x_to_i_expon */"]
	[(eok_f_to_f_expon) "/* eok_f_to_f_expon */"]
	[(eok_x_to_x_expon) "/* eok_x_to_x_expon */"]
	[(eok_remainder) "/* eok_remainder */"]
	[(eok_padd_subsc) (list " " (car ops) "[" (cadr ops) "]")]
	[(eok_pdiff) "/* eok_pdiff */"]
	[(eok_pmeq) "/* eok_pmeq */"]
	[(eok_pmne) "/* eok_pmne */"]
	[(eok_sassign) "/* eok_sassign */"]
	[(eok_bassign) "/* eok_bassign */"]
	[(eok_pmassign) "/* eok_pmassign */"]
	;[(eok_iadd_assign) "/* eok_iadd_assign */"]
	;[(eok_isubtract_assign) "/* eok_isubtract_assign */"]
	[(eok_imultiply_assign) "/* eok_imultiply_assign */"]
	[(eok_idivide_assign) "/* eok_idivide_assign */"]
	[(eok_remainder_assign) "/* eok_remainder_assign */"]
	[(eok_iadd_assign eok_fadd_assign eok_padd_assign) (list
							    (if (eqv? 'enk_variable
								      (expression-kind (address-ref (expression-operands op-exp))))
								"*"
								'()) ;HACK!!
							    (handle-lvalue (a-r (expression-operands op-exp))) " += " (cadr ops))]
	[(eok_isubtract_assign eok_fsubtract_assign eok_psubtract_assign) 
	 (list (handle-lvalue (a-r (expression-operands op-exp))) " -= " (cadr ops))]
	[(eok_fmultiply_assign) 
	 (list (handle-lvalue (a-r (expression-operands op-exp))) " *= " (cadr ops))]
	[(eok_fdivide_assign) 
	 (list (handle-lvalue (a-r (expression-operands op-exp))) " /= " (cadr ops))]
	;[(eok_padd_assign) (list (car ops) " += " (cadr ops))]
	;[(eok_psubtract_assign) "/* eok_psubtract_assign */"]
	[(eok_shiftl_assign) 
	 (list (handle-lvalue (a-r (expression-operands op-exp))) " <<= " (cadr ops))]
	[(eok_shiftr_assign) 
	 (list (handle-lvalue (a-r (expression-operands op-exp))) " >>= " (cadr ops))]
	[(eok_and_assign) "/* eok_and_assign */"]
	[(eok_or_assign) "/* eok_or_assign */"]
	[(eok_xor_assign) "/* eok_xor_assign */"]
	[(eok_subscript) (list (car ops) "[" (cadr ops) "]")]
	[(eok_field) ;;"/* eok_field */"]
	 (list (cadar ops)"." (cdr ops))]
	[(eok_value_field) "/* eok_value_field */"]
	[(eok_bit_field) "/* eok_bit_field */"]
	[(eok_value_bit_field) "/* eok_value_bit_field */"]
	[(eok_extract_bit_field) "/* eok_extract_bit_field */"]
	[(eok_pm_field) "/* eok_pm_field */"]
	[(eok_points_to_static) "/* eok_points_to_static */"]
	[(eok_lvalue_dot_static) "/* eok_lvalue_dot_static */"]
	[(eok_rvalue_dot_static) "/* eok_rvalue_dot_static */"]
	[(eok_shiftl) (list "(" (car ops) "<<" (cadr ops) ")")]
	[(eok_shiftr) (list "(" (car ops) ">>" (cadr ops) ")") ]
	[(eok_and) (list "(" (car ops) " & " (cadr ops) ")")]
	[(eok_or) "/* eok_or */"]
	[(eok_xor) "/* eok_xor */"]
	[(eok_comma) (list (car ops) (if (null? (cdr ops)) '() `(", " ,(cadr ops))))]

	[(eok_virtual_function_ptr) "/* eok_virtual_function_ptr */"]
	[(eok_vacuous_destructor_call) "/* eok_vacuous_destructor_call */"]
	[(eok_value_vacuous_destructor_call) "/* eok_value_vacuous_destructor_call */"]
	[(eok_land) (list (car ops) " & " (cadr ops))]
	[(eok_lor) (list (car ops) " | " (cadr ops))]
	[(eok_neqv) "/* eok_neqv */"]
	[(eok_eqv) "/* eok_eqv */"]
	[(eok_question) (list "(" (car ops) ")" " ? " "(" (cadr ops) ")" " : " "(" (caddr ops) ")")]
	[(eok_substring) "/* eok_substring */"]
	[(eok_value_substring) "/* eok_value_substring */"]
	[(eok_call) (if (equal? (car ops) "_main") ; HACK! by Brooke
			"/*removed call to _main*/"
			(if (equal? (car ops) "Opie415MatrixPragmaRoutine")  ; HACK! by Steve
			    "/*removed call to pragma routine generated by intermediate step of compiler*/"
			    (list (car ops) "( "
				  (if (null? (cdr ops))
				      '()
				      (cons (cadr ops)(map (lambda (x)
							     (list ", " x))
							   (cddr ops)))) " )")))]
	[(eok_virtual_call) "/* eok_virtual_call */"]
	[(eok_pm_call) "/* eok_pm_call */"]
	[(eok_fsubscript) "/* eok_fsubscript */"]
	[(eok_value_fsubscript) "/* eok_value_fsubscript */"]
	[(eok_va_start) "/* eok_va_start */"]
	[(eok_va_arg) "/* eok_va_arg */"]
	[(eok_va_end) "/* eok_va_end */"]
	[(eok_va_copy) "/* eok_va_copy */"]
	[(eok_negate) "/* eok_negate */"]
	[(eok_post_incr) "/* eok_post_incr */"]
	[(eok_post_decr) "/* eok_post_decr */"]
	[(eok_pre_incr) "/* eok_pre_incr */"]
	[(eok_pre_decr) "/* eok_pre_decr */"]
	[(eok_add) "/* eok_add */"]
	[(eok_subtract) "/* eok_subtract */"]
	[(eok_multiply) "/* eok_multiply */"]
	[(eok_divide) "/* eok_divide */"]
	[(eok_eq) "/* eok_eq */"]
	[(eok_ne) "/* eok_ne */"]
	[(eok_gt) "/* eok_gt */"]
	[(eok_lt) "/* eok_lt */"]
	[(eok_ge) "/* eok_ge */"]
	[(eok_le) "/* eok_le */"]
	[(eok_assign) "/* eok_assign */"]
	[(eok_add_assign) "/* eok_add_assign */"]
	[(eok_subtract_assign) "/* eok_subtract_assign */"]
	[(eok_multiply_assign) "/* eok_multiply_assign */"]
	[(eok_divide_assign) "/* eok_divide_assign */"]
	[(eok_address) "/* eok_address */"]
	[(eok_pm_dot_field) "/* eok_pm_dot_field */"]
	[(eok_pm_arrow_field) "/* eok_pm_arrow_field */"]
	[(eok_static_cast) "/* eok_static_cast */"]
	[(eok_const_cast) "/* eok_const_cast */"]
	[(eok_reinterpret_cast) "/* eok_reinterpret_cast */"]
	[(eok_lvalue) "/* eok_lvalue */"]
	[(eok_rvalue) "/* eok_rvalue */"]
	[(eok_generic_call) "/* eok_generic_call */"]
	
	; custom operations that use macros from dilate.h
	[(eok_even_dilate) (list "evenDilate( " (car ops) ")")]
	[(eok_odd_dilate)  (list "oddDilate( "  (car ops) ")")]
	[(eok_even_add)    (list "evenAdd( " (car ops) ", " (cadr ops) ")")]
	[(eok_odd_add)     (list "oddAdd( "  (car ops) ", " (cadr ops) ")")]
	[(eok_even_sub)    (list "evenSub( " (car ops) ", " (cadr ops) ")")]
	[(eok_odd_sub)     (list "oddSub( "  (car ops) ", " (cadr ops) ")")]
	[(eok_even_inc)    (list "evenInc( " (car ops) ")")]
	[(eok_odd_inc)     (list "oddInc( "  (car ops) ")")]
	[(eok_even_to_odd) (list "evenToOdd( " (car ops) ")")]
	[(eok_odd_to_even) (list "oddToEven( " (car ops) ")")]
	[(eok_even_add_assign) (list "(evenAddAssign(" (car ops) ", " (cadr ops) "))")]
	[(eok_odd_add_assign)  (list "(oddAddAssign( " (car ops) ", " (cadr ops) "))")]
	[(eok_my_one_const) (list "1")]
	[(eok_even_decr)    (list "evenDec( " (car ops) ")")]
	[(eok_odd_decr )   (list "oddDec( " (car ops) ")")]

	[(eok_error) "/* eok_error */"]
	[(~) "/* uninitialized op kind */"]
	[else (error 'handle-operation "Unsupported operation kind ~s" k)]))))

(define handle-constant-ref
  (lambda (cr)
    (if (null? (address-ref cr))
	"/* null constant ref */"
	(handle-constant (address-ref cr)))))

(define const->string
  (lambda (c)
;   (printf "entered const->string with: ~s~n" c)
    (cond
;****************************************************************************************
; CHANGE MADE HERE
;****************************************************************************************
;;     [(symbol? c) (list "\'" (symbol->string c) "\'")]
     [(symbol? c) (list (symbol->string c))]
;****************************************************************************************
     [(number? c) (number->string c)]
     [else (const->string (eval c))])))

; for now I'm just outputting what kind of constant it is
; you won't be able to run the code, but it should help give an idea of what is happening
; (define handle-constant
;   (lambda (c)
;     (symbol->string (constant-kind c))))

; ck_string isn't working correctly yet. I need to adjust how I'm
; reading in strings or have Steve process them before I get them
;I'm not sure if I'm handling ck_address or ck_aggregate properly
; (In fact, I think ck_aggregate is almost certainly wrong and
; ck_address is incomplete)

(define handle-constant
  (lambda (c)
    ;(printf "entered handle-constant with: ~s~n" (constant-kind c))
    (case (constant-kind c)
      [(ck_error) "/* ck_error */"]
      [(ck_integer) 	
       (let* ([val (integer_constant_tail-integer_value
		    (constant-constant_tail c))]
	      [resultval (const->string val)])
					;(printf "got back: ~s~n" resultval)
	 
	 resultval)]
      [(ck_string) (begin ;(printf "~s~n" (string_constant_tail-value (constant-constant_tail c)))
	(list "\""
			 (string_constant_tail-value (constant-constant_tail c))
			 "\""))]
      [(ck_float) (const->string (car (float_constant_tail-float_value (constant-constant_tail c))))]
      [(ck_address) (let ([act (constant-constant_tail c)])
		      (case (address_constant_tail-address_kind act)
			[(abk_constant) (handle-constant-ref (address_constant_tail-address act))]
			[(abk_variable) (list "&" (handle-variable-ref (address_constant_tail-address act)))]
			[else "/* ck_address */"]))]
      [(ck_ptr_to_member) "/* ck_ptr_to_member */"]
      [(ck_stack_offset) "/* ck_stack_offset */"]
      [(ck_dynamic_init) "/* ck_dynamic_init */"]
      [(ck_aggregate) ;"/* ck_agregate */"]

;***********************************************************************************************
; CHANGE MADE HERE
;***********************************************************************************************
       (let* ([act (constant-constant_tail c)]
	      [firstc (a-r (aggregate_constant_tail-first_constant act))])
	 ;(printf "~n~nfirstc: ~s~n~n" firstc)
	 ;(printf "~n~ncond: ~s~n~n" (a-r (constant-next firstc)))
	 (if (not (nil? (a-r (constant-next firstc))))
	     (list "{ " (handle-ck_aggregate firstc) " }")
	     (list "{ " (handle-constant firstc) " }" (string #\newline))))]
					;(list "{" (a-r (aggregate_constant_tail-first_constant act)) ", "
					;      (a-r
					;      (aggregate_constant_tail-last_constant act)) "}"))]
;*************************************************************************************************

      [(ck_init_repeat) "/* ck_init_repeat */"]
      [(ck_designator) "/* ck_designator */"]
      [(ck_template_param) "/* ck_template_param */"]
      [else (error 'handle-constant "Unsupported constant kind ~s" (constant-kind c))])))

;***********************************************************************************************
;CHANGE MADE HERE
;***********************************************************************************************
(define handle-ck_aggregate
  (lambda (const)
    ;(printf "~n~nconst: ~s~n~n" const)
    (let ([hasnext? (not (nil? (a-r (constant-next const))))])
	  (if hasnext?
	      (list " " (handle-constant const) " ," (handle-ck_aggregate (a-r (constant-next const))))
	      (list (handle-constant const))))))
;***********************************************************************************************

(define handle-variable-ref
  (lambda (vr)
    (if (null? (address-ref vr))
	"/* null var ref */"
	(handle-variable (address-ref vr)))))

;(define handle-variable_decl
;  (lambda (v)
;    (list (handle-var_storage_class v) (symbol->string (variable-name v)))))

(define handle-variable
  (lambda (v)
    (variable-name v)))
