;; -----------------------------------------------------------------------
 ; Indiana University Opie Lab Software License, Version 1.2
 ;
 ; Copyright (C) 2002 The Trustees of Indiana University.
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without
 ; modification, are permitted provided that the following conditions are
 ; met:
 ;
 ; 1) All redistributions of source code must retain the above
 ;    copyright notice, the list of authors in the original source
 ;    code, this list of conditions and the disclaimer listed in this
 ;    license;
 ;
 ; 2) All redistributions in binary form must reproduce the above
 ;    copyright notice, this list of conditions and the disclaimer
 ;    listed in this license in the documentation and/or other
 ;    materials provided with the distribution;
 ;
 ; 3) Any documentation included with all redistributions must include
 ;    the following acknowledgement:
 ;
 ;      "This product includes software developed by the Indiana
 ;      University Opie Lab.  For further information please visit
 ;      http://www.cs.indiana.edu/~dswise"
 ;
 ;    Alternatively, this acknowledgment may appear in the software
 ;    itself, and wherever such third-party acknowledgments normally
 ;    appear.
 ;
 ; 4) The name "Indiana University" or "Indiana University
 ;    Opie Lab" shall not be used to endorse or promote
 ;    products derived from this software without prior written
 ;    permission from Indiana University.  For written permission,
 ;    please contact http://www.cs.indiana.edu/~dswise.
 ;
 ; 5) Products derived from this software may not use "Indiana
 ;    Univeristy" name nor may "Indiana University" appear in their name,
 ;    without prior written permission of the Indiana University.
 ;
 ; Indiana University provides no reassurances that the source code
 ; provided does not infringe the patent or any other intellectual
 ; property rights of any other entity.  Indiana University disclaims any
 ; liability to any recipient for claims brought by any other entity
 ; based on infringement of intellectual property rights or otherwise.
 ;
 ; LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH
 ; NO WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA
 ; UNIVERSITY GIVES NO WARRANTIES AND MAKES NO REPRESENTATION THAT
 ; SOFTWARE IS FREE OF INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR
 ; OTHER PROPRIETARY RIGHTS.  INDIANA UNIVERSITY MAKES NO WARRANTIES THAT
 ; SOFTWARE IS FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP
 ; DOORS", "WORMS", OR OTHER HARMFUL CODE.  LICENSEE ASSUMES THE ENTIRE
 ; RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR ASSOCIATED MATERIALS,
 ; AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION GENERATED USING
 ; SOFTWARE.
;; ------------------------------------------------------------------------

;(load "match.ss")
;(load "scil-records.ss")
;(load "scil-parse.ss")
;(load "scil-to-c.ss")
;(load "scil-utilities.ss")
;(load "gen-utilities.ss")

;
; define-syntax
; syntax-case
; show-paren-mode
;
; Known bugs:
;   array declaration ordering 
;   -- FIXED June '04 mrcarlso

;   shadow-assignment  
;   storage classes
;   external function decl's
;   
;   dynamic inits DO need to go in 
;   -- supports init to constants June '04 mrcarlso  
;


(define apply-to-statement-kind
  (lambda (stmts mregion f)
    (cond
     [(null? stmts) #f]
     [(statement-deleted (car stmts)) (apply-to-statement-kind (cdr stmts) f)]
     [(statement-kind (car stmts)) 
      (begin
	(f (car stmts) mregion)
	(apply-to-statement-kind (cdr stmts) f)
	#t)]
     [else (apply-to-statement-kind (cdr stmts) f)])))


; (define fix-for-initialization
;   (lambda (stmt mregion)
;     (printf "fixing init on ~s~n" stmt)
;     (if (nil? (a-r (statement-initialization stmt)))
; 	stmt
; 	(let ([last-stmt (find-last-stmt-in-list (a-r (statement-initialization stmt)))]
; 	      [ptr (a-r (statement-initialization stmt))])
; 	  (printf "init is ~s~n" ptr)
; 	  (change-statement last-stmt `((next ,ptr)))
; 	  (change-statement stmt `((init ())))
; 	  (swap-statements stmt (a-r ptr))
; 	  ;(change-statement ptr `((next ,(a-r ptr))))
; 	  ptr))))

(define fix-for-initialization
  (lambda (stmt mregion)
    (let ([init (a-r (statement-initialization stmt))])
      (change-statement stmt `((init ())))
      (insert-before-statement stmt init))))
     
(define fix-for-test
  (lambda (stmt mregion)
    (let* ([first-stmt  (a-r (statement-statements (a-r (statement-statement stmt))))]
	   [new-if-stmt (new-if 
			 (construct-expression mregion
					       `(kind enk_operation)
					       `(opkind eok_not)
					       `(operands ,(a-r (statement-expr stmt))))
			 (construct-statement mregion `(kind stmk_mybreak))
			 `()
			 first-stmt
			 mregion)])
      (insert-before-statement first-stmt new-if-stmt)
      (change-statement stmt `((expr ,(construct-expression mregion 
							    `(kind enk_operation)
							    `(opkind eok_my_one_const))))))))

; (define fix-for-test
;   (lambda (stmt mregion)
;     (let* ([[first-stmt  (a-r (statement-statements (a-r (statement-statement stmt))))]
; 	    [new-if-stmt (new-if 
; 			  (construct-expression mregion
; 						`(kind enk_operation)
; 						`(opkind eok_not)
; 						`(operands ,(a-r (statement-expr stmt))))
; 			  (construct-statement mregion `(kind stmk_mybreak))
; 			  `()
; 			  first-stmt
; 			  mregion)])
; 	   (change-statement stmt `((expr 

(define fix-for-increment
  (lambda (stmt mregion)
    (let ([last-stmt (find-last-stmt-in-block (a-r (statement-statement stmt)))]
	  [new (exprs->stmts (a-r (statement-increment stmt)) mregion)])
      
      (change-statement last-stmt `((next ,new)))
      (change-statement stmt `((incr ()))))))
	       

(define fix-add-assign-stmt
  (lambda (stmt mregion)
    (cond
     [(null? stmt) #f]
     [(not (and (statement? stmt) 
		(equal? (statement-kind stmt) 'stmk_expr)
		(equal? (expression-kind (a-r (statement-expr stmt))) 'enk_operation)
		(equal? (expression-operation_kind (a-r (statement-expr stmt))) 'eok_iadd_assign)))
      #f]
     [else
      (let ([copy-of-target (var_addr_expr->var_expr (a-r (expression-operands (a-r (statement-expr stmt)))) mregion)])
	(change-expression (a-r (statement-expr stmt)) `((opkind eok_iassign)))
	(set-next (a-r (expression-operands (a-r (statement-expr stmt))))
		  (new-op 
		   'eok_iadd 
		   '() 
		   mregion 
		   copy-of-target
		   (a-r (expression-next (a-r (expression-operands (a-r (statement-expr stmt))))))))
	#t)])))

(define set-next
  (lambda (target next)
    (cond
     [(statement? target) 
      (if (not (statement? next)) 
	  (error 'set-next "target is a statement, next isn't~n")
	  (change-statement target `((next ,next))))]
     [(expression? target) 
      (if (not (expression? next))
	  (error 'set-next "target is an expression, next isn't~n")
	  (change-expression target `((next ,next))))]
     [else (error 'set-next "invalid target~n")])))

(define get-next
  (lambda (target)
    (cond
     [(statement? target) (a-r (statement-next stmt))]
     [(expression? target) (a-r (expression-next stmt))])))

(define stmt-has-matrix-variable?
  (lambda (stmt mregion)
    (letrec ([search-matrix-info
              (lambda (stmt mtrx-info)
                (cond
                 [(null? stmt) '()]
                 [(null? matrix-info) '()]
                 [(statement-contains-var? stmt (caar matrix-info))
                  (cons (caar matrix-info) 
                        (search-through-matrix-variables stmt (cdr matrix-info))
                  )
                 ]
                 [else (search-through-matrix-variables stmt (cdr matrix-info))]
                )
              )
            ])
      (search-matrix-info stmt (mregion-matrix_info))
    )
  )
)
		  
;
; Copies stmtb into stmta 
;
(define copy-statement 
  (lambda (stmta stmtb)
    (set-statement-number! stmta (statement-number stmtb))
    (set-statement-kind! stmta (statement-kind stmtb))
    (set-statement-position! stmta (statement-position stmtb))
    (set-statement-final_position! stmta (statement-final_position stmtb))
    (set-statement-expr! stmta (statement-expr stmtb))
    (set-statement-statements! stmta (statement-statements stmtb))
    (set-statement-assoc_scope! stmta (statement-assoc_scope stmtb))
    (set-statement-lifetime! stmta (statement-lifetime stmtb))
    (set-statement-end_of_block_reachable! stmta (statement-end_of_block_reachable stmtb))
    (set-statement-then_statement! stmta (statement-then_statement stmtb))
    (set-statement-else_statement! stmta (statement-else_statement stmtb))
    (set-statement-initialization! stmta (statement-initialization stmtb))
    (set-statement-statement! stmta (statement-statement stmtb))
    (set-statement-increment! stmta (statement-increment stmtb))
    (set-statement-has_empty_else_clause! stmta (statement-has_empty_else_clause stmtb))
    (set-statement-dynamic_init! stmta (statement-dynamic_init stmtb))
    (set-statement-deleted! stmta (statement-deleted stmtb)))) 

(define copy-expression
  (lambda (expra exprb)
    (set-expression-number! expra (expression-number exprb))
    (set-expression-next! expra (expression-next exprb))
    (set-expression-kind! expra (expression-kind exprb))
    (set-expression-operation_kind! expra (expression-operation_kind exprb))
    (set-expression-routine! expra (expression-routine exprb))
    (set-expression-type! expra (expression-type exprb))
    (set-expression-operands! expra (expression-operands exprb))
    (set-expression-variable! expra (expression-variable exprb))
    (set-expression-constant! expra (expression-constant exprb))
    (set-expression-result_is_not_used! expra (expression-result_is_not_used exprb))
    (set-expression-compiler_generated! expra (expression-compiler_generated exprb))
    (set-expression-field! expra (expression-field exprb))
    (set-expression-odd_typing! expra (expression-odd_typing exprb))
    (set-expression-even_typing! expra (expression-even_typing exprb))
    (set-expression-odd_terms! expra (expression-odd_terms exprb))
    (set-expression-even_terms! expra (expression-even_terms exprb))
    (set-expression-updated! expra (expression-updated exprb))))
;
; A temporary statement variable, used to swap other statements.
;
(define temp-statement
  (make-statement 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0))

(define temp-expr
  (make-expression 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 #f))

(define swap-statements
  (lambda (stmta stmtb)
    ;(printf "swapping ~s~nand~n~s~n" stmta stmtb)
    (copy-statement temp-statement stmta)
    (copy-statement stmta stmtb)
    (copy-statement stmtb temp-statement)))

(define swap-expressions
  (lambda (expra exprb)
    (copy-expression temp-expr expra)
    (copy-expression expra exprb)
    (copy-expression exprb temp-expr)))

;
; This function will get the next statement, being smart and avoiding
; statements that have been deleted.
;
(define next-stmt
  (lambda (stmt)
    (let ([nexts (address-ref (statement-next stmt))])
      (cond
       [(statement-deleted next) (next-stmt nexts)]
       [else nexts]
      )
    )
  )
)

(define add-after-statement
  (lambda (stmt add_stmt mregion)	
    (set-address-ref! (statement-next add_stmt) (address-ref (statement-next stmt)))
    (set-address-ref! (statement-next stmt) add_stmt)
  )
)


(define add-before-statement
  (lambda (stmt add_stmt mregion)
    '()   
  )
)        


; Deletes a statement by setting its deleted flagged to true.
; 
; NOTE: nothing else happens.  If you want other statements to stop
;       linking to do this, then make sure you do that yourself!
;
(define delete-statement
  (lambda (statement)
    (set-statement-deleted! statement #t)
  )
)

;
; This function searches a memory region, looking for any calls to the
; pragma function.  When/if the pragma function is called, that statement
; is deleted and pragma information is injected into the mregion.
;
; NOTE: right now I'm searching exprs, which is bad because I don't
;       actually delete the function call, FIX THIS.
;

; (define find-pragmas
;  (lambda (mregion)
;   (letrec
;     ;
;     ; This helper function takes an expression that should be the first
;     ; argument of an eok_call operation.  It looks at the second and
;     ; third arguments and returns a pair containing two pairs.  The first
;     ; pair is a variable name and variable of a matrix.  The second pair
;     ; is the variable name and variable of that matrix's stride.
;     ; 
;     ([make-matrix-entry
;       (lambda (expr) 
;        (cond
;         [(null? (expression-next expr)) 
;          (error 'find-pragmas "not enough arguments to pragma function~n")
;         ]
;         [(null? (expression-next (a-r (expression-next expr))))
;          (error 'find-pragmas "not enough arguments to pragma function~n")
;         ]
;         [else 
;          (let ([mexpr (strip-casts (a-r (expression-next expr)))]
;                [sexpr (strip-casts (a-r (expression-next (a-r (expression-next expr)))))]
;               )
	   
; 	   (if 
; 	    (null? (expression-variable (strip-casts mexpr)))   
; 	    (error 'find-pragmas "arguments to pragma variable must be simple variables~n")
;            (if
;             (null? (expression-variable (strip-casts sexpr)))
;             (error 'find-pragmas "arguments to pragma variable must be simple variables~n")
; 	    (if (equal? (expression-kind sexpr) 'enk_variable)
; 		(make-one_matrix          
; 		 (a-r (expression-variable mexpr))
; 		 `(var ,(a-r (expression-variable sexpr)))
; 		 '()
; 		 '()
; 		 '())
; 		(if (and (equal? (expression-kind sexpr) 'enk_operation)
; 			 (equal? (expression-operation_kind sexpr) 'eok_indirect))
; 		    (make-one_matrix
; 		     (a-r (expression-variable mexpr))
; 		     `(ptr ,(a-r (expression-variable (a-r (expression-operands sexpr)))))
; 		     '()
; 		     '()
; 		     '())
; 		    (error 'find-pragmas "cannot understand stride argument to Opie pragma (operation kind: ~s)" (expression-operation_kind sexpr)))
; 		    ))))]))]
;      ;
;      ; This helper function takes a list of expressions and searches them for
;      ; any eok_calls of the pragma function.
;      ;
;      [search-exprs-for-pragmas 
;        (lambda (exprs) 
;         (cond
;          [(null? exprs) '()]
;          [(and
;            (equal? (expression-kind (car exprs)) 'enk_operation)
;            (equal? (expression-operation_kind (car exprs)) 'eok_call))
;           (cond 
;            [(null? (expression-operands (car exprs))) 
;             (search-exprs-for-pragmas (cdr exprs))]
;            [(null? (expression-routine (a-r(expression-operands (car exprs)))))
;             (search-exprs-for-pragmas (cdr exprs))]
;            [(equal? 
;              (routine-unmangled_name (a-r (expression-routine 
;               (a-r(expression-operands (car exprs))))))                         
;              'Opie415MatrixPragmaRoutine)
; 	    ;(printf "making matrix entry with ~s~n" (expression-operation_kind (strip-casts (a-r (expression-next (a-r (expression-next (a-r (expression-operands (car exprs))))))))))
;             (cons 
;              (make-matrix-entry (a-r (expression-operands (car exprs))))
;              (search-exprs-for-pragmas (cdr exprs))
;             )
;            ]
;            [else (search-exprs-for-pragmas (cdr exprs))]
;           )
;          ]
;          [else (search-exprs-for-pragmas (cdr exprs))]                 
;         ) 
;        )
;     ])   
;    (search-exprs-for-pragmas (memory_region-expression_list mregion))
;   )
;  )
; )

;
; Search through a memory region looking for matrix dereferences
; and return a list of pairings of variables with expressions that
; dereference them.  Each variable may be mentioned multiple times
; if it is dereferenced multiple times.
;
(define find-matrix-dereferences
  (lambda (mregion)
    (letrec ([search-exprs-for-matrix-deref
              (lambda (exprs matrix_list)
		;(printf "entered search-exprs-for-matrix-deref with exprs:~n~s~n" exprs)
		;(printf "null? exprs: ~s~n" (null? exprs))
                (cond
                 [(null? exprs) '()]
                 [(null? matrix_list) '()]
                 [(or (equal? (expression-operation_kind (car exprs)) 'eok_padd_subsc)
                      (equal? (expression-operation_kind (car exprs)) 'eok_padd)
		      (equal? (expression-operation_kind (car exprs)) 'eok_subscript))
                  ;(printf "found a padd expr~n")
                  (if (mempred
		       (a-r (expression-variable (strip-casts (a-r (expression-operands (car exprs))))))
		       matrix_list)
                      (cons (list (a-r (expression-variable (strip-casts (a-r (expression-operands (car exprs))))))
				  (a-r (expression-next (a-r (expression-operands (car exprs))))))
			    (search-exprs-for-matrix-deref (cdr exprs) matrix_list))
		      (search-exprs-for-matrix-deref (cdr exprs) matrix_list))]
                 [else (search-exprs-for-matrix-deref (cdr exprs) matrix_list)]))])
      (search-exprs-for-matrix-deref (memory_region-expression_list mregion) 
                                     (map one_matrix-variable (memory_region-matrix_info mregion))))))
 
;
; Function that sets the stride portion of the matrix info stored for
; a memory region, using find-pragmas.
;
(define set-matrix-info-strides
  (lambda (mregion)
    (set-memory_region-matrix_info! mregion (find-pragmas mregion)) mregion))

;
; Takes a list and returns a sublist that contains only those elements
; of the original that had a car matching the paramater "match" used
; by set-matrix-info-derefs
;
(define get-only-car-matches
  (lambda (ls match)
    (cond
     [(null? ls) '()]
     [(equal? (caar ls) match)
      (cons (car ls) (get-only-car-matches (cdr ls) match))
     ]
     [else (get-only-car-matches (cdr ls) match)]
    )
  )
)

(define find-matrix
  (lambda (var matrices)
    (cond
     [(null? matrices) (error 'find-matrix "matrix not found")]
     [(eq? var (one_matrix-variable (car matrices))) (car matrices)]
     [(find-matrix var (cdr matrices))])))

;
; Uses find-matrix-dereferences and get-only-car-matches to find all
; matrix dereferences for matrices that we have pragma info on, and
; store the locations thereof in the memory region.  Should be called
; AFTER pragma information has been read with set-matrix-info-strides.
(define set-matrix-info-derefs
  (lambda (mregion)
    (let* ([derefs (find-matrix-dereferences mregion)]
	   [vars (map car derefs)]
	   [exprs (map cadr derefs)])
      (define helper
	(lambda (vars exprs mregion)
	  ;(printf "entered helper with vars:~n~s~n" vars)
	  ;(printf "entered helper with exprs:~n~s~n" exprs)
	  (cond
	   [(null? vars) '()]
	   [else 
	    (let ([matrix (find-matrix (car vars) (memory_region-matrix_info mregion))])
	      (set-one_matrix-deref_exprs! 
	       matrix 
	       (cons (car exprs)
		     (one_matrix-deref_exprs matrix)))
	      (helper (cdr vars) (cdr exprs) mregion))])))
      (helper vars exprs mregion)
      ;(printf "we did it~n")
      )))

; (define set-matrix-info-derefs
;   (lambda (mregion)
;     (letrec ([set-one-matrix-derefs
;                (lambda (matrix-info matrix-derefs)
; 		 (printf "entered set-one-matrix-derefs with matrix-info:~n~s~n" matrix-info)
;                  (set-one_matrix-deref_exprs! 
;                    (car matrix-info) 
;                    (map cadr (get-only-car-matches matrix-derefs (one_matrix-variable (car matrix-info))))
;                  )
;                )
;            ])
;       (set-one-matrix-derefs 
;          (memory_region-matrix_info mregion)  
;          (find-matrix-dereferences mregion)
;       )
;       mregion
;     )
;   )
; )

;
; Retrieves from a memory region, the information stored regarding
; one pragma'ed matrix.
;
(define get-matrix-entry
  (lambda (matrixvar ls)
    (cond
     [(null? ls) (error 'get-matrix-entry "matrix not found~n")]
     [(equal? matrixvar (one_matrix-variable (car ls))) (car ls)]
     [else (get-matrix-entry (car ls))]
    )
  )
)

;
; Add an odd variable to the listing of odd variables in the matrix
; info stored in a memory region.
;
(define add-odd-var
  (lambda (var matrixvar matrixinfo)
    (if (mempred var (one_matrix-odd_vars matrix-entry))
        #f
        (begin
          (set-one_matrix-odd_vars! matrix-entry (cons var (one_matrix-odd_vars matrix-entry)))
          #t
        )
      
    )
  )
)

;
; As above for even variables.
;
(define add-even-var
  (lambda (var matrixvar matrixinfo)
    (if (mempred var (one_matrix-even_vars matrix-entry))
        #f
        (begin
          (set-one_matrix-odd_vars! matrix-entry (cons var (one_matrix-even_vars matrix-entry)))
          #t
        )
      
    )
  )
)
      
;
; Takes  list of terms, and negates it, either by adding a 'neg_one_const
; in front of each term, or removing a 'neg_one_const from that term.
; 'neg_one_const is just a convenient way of turning addition into
; subtraction.
; 

(define negate-term
  (lambda (t)
    (match t
      [one_const 'neg_one_const]
      [neg_one_const 'one_const]
      [(neg_one_const ,xs ...) `(,xs ...)]
      [(,xs ...) `(neg_one_const ,xs ...)]
      [,x `(neg_one_const ,x)])))

(define negate-terms-list
  (lambda (xs)
      (map negate-term xs)))



; (define negate-terms-list
;   (lambda (ls)
;     (cond
;      [(null? ls) '()]
;      [(list? (car ls))
;       (if (equal? (caar ls) 'neg_one_const)
;           (cons (cdar) (negate-terms-list ls))
;           (cons (cons 'neg_one_const (car ls)) (negate-terms-list ls))
;       )
;      ]
;      [(equal? (car ls) 'neg_one_const)
;       (cons 'one_const (negate-terms-list ls))
;      ]
;      [equal? (car ls) 'one_const
;       (cons 'neg_one_const (negate-terms-list ls)) 
;      ]
;      [else (cons (list 'neg_one_const (car ls)) (negate-terms-list ls))]
;     )
;   )
; )

;
; Take two terms (lists of variables or constants) and multiply them
; together, taking car to properly handle 'neg_one_const's and
; one_const's (the rule here being that there should be only one per
; term and it should be the first element of the term.
;
(define multiply-single-terms
  (lambda (xs ys)
    (let ([x (car xs)]
          [xcdr (cdr xs)]
          [y (car ys)]
          [ycdr (cdr ys)])
      (cond
       [(equal? x 'one_const)
	(if (equal? y 'one_const)
	    (if (or (not (null? xcdr)) 
		    (not (null? ycdr))) 
		(append ycdr xcdr)
		(list 'one_const))
	    (append ycdr xcdr))]
       [(equal? x 'neg_one_const)
	(if (equal? y 'neg_one_const)
	    (if (or (not (null? xcdr)) 
		    (not (null? ycdr))) 
		(append ycdr xcdr)
		(list 'one_const))
	    (if (equal? y 'one_const)
		(append xs ycdr)
		(append xs ys)))]
       [else (if (equal? y 'one_const)
		 (append xs ycdr)
		 (append ys xs))]))))

(define singleton?
  (lambda (x)
    (not (pair? x))))

(define multiply-terms
  (lambda (xs ys)
    ;(printf "entered multiply-terms with:~n~s~n~s~n" xs ys)
    (let ()
      (define multiply-term/terms
	(lambda (x ys)
	  ;(printf "entered multiply-term/terms with:~n~s~n~s~n" x ys)
	  (match x
	    [(one_const) ys]
	    [(neg_one_const zs ...) (map (lambda (y) 
					   (if (equal? (car y) 'neg_one_const) 
					       (append (cdr `(,zs ...)) (cdr y)) 
					       (append `(neg_one_const ,zs ...) y)))
					 ys)]
	    [(,zs ...) (map (lambda (y) (append y `(,zs ...))) ys)]						 
	    [,otherwise (error 'multiply-term/terms "non-list element of a term")])))
      (flatten-once (map (lambda (x) (multiply-term/terms x ys)) xs)))))

;
; Takes two lists of terms (see def'n above) and multiplies them.
; This is basically the multiplication of two additive structures.
; Example:
;  ((a b) 1) * ((c) (-1 d))
; is a hacked up way of doing
;  (a*b+1) * (c-d)
; and will return:
;  ((a b c) (-1 a b d) (c) (-1 d))
; where 1 and -1 are one_const and neg_one_const (if they are actual
; constants put in by the user, that's fine, but we can't handle clean
; them up so nice and will assume that cc does this for us) and the
; variable names are really references to expressions that are of kind
; variable or constant.  Terms can also contain certain operations,
; and even functions, but these won't all supported yet.
;
(define multiply-multiple-terms
  (lambda (xs ys) 
    (cond
     [(null? xs) '()]
     [else (append (map (lambda (x) (multiply-single-terms x ys)) xs)
		   (multiply-multiple-terms (cdr xs) ys))])))


(define var-stride?
  (lambda (stride)
    (equal? (car stride) `var)))

;
; This function drives the identification part of the compiler.
;
; It's job is to take an expression that is an interesting matrix
; dereference and tear it apart, finding out which expressions 
; (meaning variables, constants, or any shadowable functions/operations)
; to shadow as odd integers and which to shadow even.
;
; Basically, at the leaf level, everything but a stride variable
; is odd.  As the traversal climbs up the tree however, expressions
; can be "promoted" to even status through multiplication with an
; existing even reference.  Strides are converted to an even one
; constant (with 'one_const, which is a flag that will be removed
; at any stage where it is possible, i.e. after multiplication
; with any term that is not equal to 'one_const.
;
; Everything is treated as part of a multiplicative term.  A list
; of such multiplicative terms represents the addition of these
; terms.  As the traversal happens, the expression is
; reconstructed into a somewhat normalized form, of added
; multiplicative terms (no additions within such a multiplicative
; term).  Subtraction is addition of negative terms, with
; 'neg_one_const acting as a flag that indicates negativeness.
; No 'neg_one_consts will survive later stages of the compiliation
; as they will be turned back into subtractions.

(define dilate-traverse
  (lambda (expr stride)
    (let*([kind (expression-kind expr)]
	  [opkind (expression-operation_kind expr)]
	  [ops (a-r (expression-operands expr))]
	  [ops2 (if (nil? ops) '() (a-r (expression-next ops)))]
	  [var (a-r (expression-variable expr))]
	  [rtn (expression-routine expr)])
;      (printf "in dilate-traverse~nkind:~s~nop:~s~nvar:~s~n" kind opkind)
      (cond
       [(equal? kind 'enk_variable)
	(if (and (var-stride? stride) (equal? var (cadr stride)))
            (set-expression-even_terms! expr `((one_const)))
            (set-expression-odd_terms! expr `((,expr))))]

       [(equal? kind 'enk_constant)
	(set-expression-odd_terms! expr `((,expr)))]

       [(equal? kind 'enk_operation)
	(cond
	 [(and (equal? opkind 'eok_indirect)
	       (not (var-stride? stride))
	       (equal? (a-r (expression-variable (strip-casts (a-r (expression-operands expr))))) (cadr stride)))
	  (set-expression-even_terms! expr `((one_const)))]
		    
	 [(or (equal? opkind 'eok_call)
	      (equal? opkind 'eok_and)                
	      (equal? opkind 'eok_or)                
	      (equal? opkind 'eok_xor)                
	      (equal? opkind 'eok_shiftl)                
	      (equal? opkind 'eok_shiftr)
              (equal? opkind 'eok_indirect))
	  (set-expression-odd_terms! expr `((,expr)))]

	 [(equal? opkind 'eok_iadd)
	  (dilate-traverse ops stride)
	  (dilate-traverse (a-r (expression-next ops)) stride)
	  (set-expression-even_terms! 
	   expr
	   (append (expression-even_terms ops)
		   (expression-even_terms (a-r (expression-next ops)))))
	  (set-expression-odd_terms! 
	   expr
	   (append (expression-odd_terms ops)
		   (expression-odd_terms (a-r (expression-next ops)))))]

	 [(equal? opkind 'eok_isubtract)
	  (dilate-traverse ops stride)
	  (dilate-traverse (a-r (expression-next ops)) stride)
	  (set-expression-even_terms! 
	   expr
	   (append (expression-even_terms ops)
		   (negate-terms-list 
		    (expression-even_terms (a-r (expression-next ops))))))
	  (set-expression-odd_terms! 
	   expr
	   (append (expression-odd_terms ops)
		   (negate-terms-list 
		    (expression-odd_terms (a-r (expression-next ops))))))]

	 [(equal? opkind 'eok_imultiply) '()
	  (dilate-traverse ops stride)
	  (dilate-traverse ops2 stride)
	  ;(printf "in multiply~n op1 odd: ~s~n op1 even ~s~n op2 odd: ~s~n op2 even: ~s~n"
;		  (expression-odd_terms ops)
	;	  (expression-even_terms ops)
		;  (expression-odd_terms ops2)
		 ; (expression-even_terms ops2))
	  (if (and (null? (expression-even_terms  ops))        
		   (null? (expression-even_terms ops2)))
	      (set-expression-odd_terms! 
	       expr 
	       (multiply-terms (expression-odd_terms ops) (expression-odd_terms ops2)))
	      (if (null? (expression-even_terms  ops)) 
		  (begin
		    (set-expression-odd_terms!
		     expr
		     (multiply-terms (expression-odd_terms ops) (expression-odd_terms ops2)))
		    (set-expression-even_terms! 
		     expr
		     (multiply-terms (expression-odd_terms ops) (expression-even_terms ops2))))
		  (if (null? (expression-even_terms ops2))
		      (begin
			(set-expression-odd_terms! 
			 expr
			 (multiply-terms (expression-odd_terms ops) (expression-odd_terms ops2)))
			(set-expression-even_terms!
			 expr
			 (multiply-terms (expression-even_terms  ops) (expression-odd_terms ops2))))
		      (error 'dilate-traverse "two stride terms multiplied~n"))))]
           [else (error 'dilate-traverse "unhandled op ~s~n" opkind)])]
	[else (error 'dilate-traverse "unhandled expr type ~s~n" kind)]))))


(define parse-matrix-derefs
  (lambda (mregion)
    (letrec ([pmd-from-matrix-info
              (lambda (ls)
                (if (null? ls)
                    '()
                    (begin
                      (map (lambda (m) (dilate-traverse m (one_matrix-stride (car ls))))
                           (one_matrix-deref_exprs (car ls))
                      )
                      (pmd-from-matrix-info (cdr ls))
                    )
                )
              )
             ]
            )
      (pmd-from-matrix-info (memory_region-matrix_info mregion))
    )
  )
)

(define set-odd-even-from-expr
  (lambda (expr matrix mregion)
    ;(printf "entered set-odd-even-from-expr~n")
    (let ()
      (define flatten
	(lambda (ls)
	 ; (printf "entered flatten with ls:~n~s~n" ls)
	  (cond
	   [(null? ls) '()]
	   [(pair? ls) (append (flatten (car ls)) (flatten (cdr ls)))]
	   [else (list ls)])))

      (define find-vars
	(lambda (ls)
	;  (printf "entered find-vars~n")
	  (cond
	   [(null? ls) '()]
	   [(and (expression? (car ls)) 
		 (equal? (expression-kind (car ls)) 'enk_variable))
	    (cons (a-r (expression-variable (car ls))) (find-vars (cdr ls)))]
	   [else (find-vars (cdr ls))])))

      (define helper 
	(lambda (vars even? matrix mregion)
	  ;(printf "entered helper with ~s~n" vars)
	  (cond
	   [(null? vars) '()]
	   [(member? (car vars) (one_matrix-odd_vars matrix))
	    (if even?
		(begin
		  (set-one_matrix-odd_vars! matrix (remove (car vars) (one_matrix-odd_vars matrix)))
		  (set-one_matrix-even_vars! matrix (cons (car vars) (one_matrix-even_vars matrix)))))
	    (helper (cdr vars) even? matrix mregion)]
	   [(member? (car vars) (one_matrix-even_vars matrix))
	    (helper (cdr vars) even? matrix mregion)]
	   [else
	    (if even?
		(set-one_matrix-even_vars! matrix (cons (car vars) (one_matrix-even_vars matrix)))
		(set-one_matrix-odd_vars! matrix (cons (car vars) (one_matrix-odd_vars matrix))))
	    (helper (cdr vars) even? matrix mregion)])))

      ;(printf "entered setevenodd with expr:~n~s~n" expr)
      (let ([evens (find-vars (flatten (expression-even_terms expr)))]
	    [odds  (find-vars (flatten (expression-odd_terms  expr)))])
      (helper evens #t matrix mregion)
      (helper odds  #f matrix mregion)))))
     
(define set-matrix-info-oddeven
  (lambda (mregion)
    (letrec ([smioe-from-matrix-info
              (lambda (ls)
                (if (null? ls)
                    '()
                    (begin
                      (map (lambda (exprs) (set-odd-even-from-expr exprs (car ls) mregion)) (one_matrix-deref_exprs (car ls)))
                      (smioe-from-matrix-info (cdr ls)))))])
      (smioe-from-matrix-info (memory_region-matrix_info mregion)))))

;
; Takes an expression and returns a list containing it, and all "next" expressions or expressions
; that may be linked from this expression via other fields (specifically operands).
;
(define expr->exprs
  (lambda (expr)
    (if (null? expr)
	'()
	(cons 
	 expr
	 (append (if (not (null? (expression-next expr)))
		     (expr->exprs (a-r (expression-next expr))))
		 (if (not (null? (expression-operands expr)))
		     (expr->exprs (a-r (expression-operands expr)))))))))
(define flatten-once
  (lambda (ls)
    (cond
     [(null? ls) '()]
     [(append (car ls) (flatten-once (cdr ls)))])))

; expr->vars :: expr -> [var]
;
(define expr->vars
  (lambda (expr)
    (cond
     [(nil? expr) '()]
     [(equal? (expression-kind expr) 'enk_variable)
      (a-r (expression-variable expr))]
     [(equal? (expression-kind expr) 'enk_operation)
      (flatten (map expr->vars (get-list-of-records expression-next (a-r (expression-operands expr)))))]
     [else '()])))
      

; stmt->exprs :: stmt -> [expr]
;
; Takes a statement and returns all subexpressions, where subexpression is defined to be, any
; expr->exprs of any expression linked directly from the statement.  It should be the case that 
; the union of the subexpressions of all statements in an mregion, is the same as the set of all
; expressions in the mregion, and that the intersection of subexpressions of any two statements,
; is always empty.
; 
(define stmt->exprs
  (lambda (stmt)
    (append (if (and (not (null? stmt)) (not (null? (statement-expr stmt))))
		(expr->exprs (a-r (statement-expr stmt))) 
		'())
	    (if (and (not (null? stmt)) (not (null? (statement-increment stmt))))
		(expr->exprs (a-r (statement-increment stmt))) 
		'()))))

; expr->stmt :: expr -> stmt
;
; Wraps an expression in a statement.
;
; The resulting statement is very plain (in particular it points to no next stmt).
;
(define expr->stmt
  (lambda (expr mregion)
    (construct-statement
     mregion
     `(next ())
     `(kind stmk_expr)
     `(expr ,expr))))

; exprs->stmt :: expr -> stmt
;
; Takes a list of expressions (list in the IL sense of list) and converts it into
; a list of statements using expr->stmt (again, in the IL sense of a list), with
; the last statement pointing to no "next" statement.
;
(define exprs->stmts 
  (lambda (exprs mregion)
    ;(printf "~nexprs->stmts with: ~s~n" exprs)
    (cond
     [(nil? exprs) '()]
     [else
      (let ([new (expr->stmt exprs mregion)]
	    [next (a-r (expression-next exprs))])
	(change-expression exprs `((next ())))
	(change-statement new `((next ,(exprs->stmts next mregion))))
	new)])))
	

;
; ormap-return takes a predicate, and a list, and returns the first element of the list, for
; which the predicate is true when applied to it.  If no element of the list is true under
; the predicate then #f is returned.
;
(define ormap-returns
  (lambda (pred xs)
    (match xs
      [() #f]
      [(,a . ,as) (guard (pred a)) a]
      [(,a . ,[as]) as])))

(define new-address
  (lambda (target)
    (let ([new (make-address '() '() '())])
      (set-address-ref! new target)
      new)))

;
; Finds the highest statement number in a memory region and returns it.
;
(define get-highest-statement
  (lambda (mregion)
    (+ 1 (length (memory_region-statement_list mregion)))))

;
; Makes a new statement and returns it.  Requiresa memory region as an
; argument.  It will set the statement # correctly, and add the
; statement to the memory region's statement_list.
;
(define new-statement
  (lambda (mregion)
    (if 
      (null? mregion)
      (error 'new-statement "new-statement must take a memory region as an argument~n"))
    (let ([new (make-statement 
		(get-highest-statement mregion) 0 'stmk_error 0 0 0 0 0
		0                               0 0           0 0 0 0 0 0 0 #f)])
      (set-memory_region-statement_list! 
       mregion 
       (append (memory_region-statement_list mregion) (list new)))
      new)))


(define get-highest-expression
  (lambda (mregion)
    (+ 1 (length (memory_region-expression_list mregion)))))

(define new-expression
  (lambda (mregion)
    (if 
      (null? mregion)
      (error 'new-expression "new-expression must take a memory region as an argument~n"))
    (let ([new (make-expression
		(get-highest-expression mregion) 0  'enk_error 0 0 0 0 0
		 0                               #f 0          0 0 0 0 0 #f)])
      (set-memory_region-expression_list! 
      mregion 
       (append (memory_region-expression_list mregion) (list new)))
      new)))

;
; This is a more friendly mechanism for modifying a statement.  It takes
; a statement, and an arbitrarily long list of fields and new entries
; for these fields.  fields that should take an addressrecord, should be
; passed a reference and a new address will be created for that.
;
(define change-statement
  (lambda (stmt args)
;    (printf "entered change statement with args: ~s~n" args)
    (match args
      [() stmt]
      [((next ,s) ,rest ...)
       (set-statement-next! stmt (new-address s))
       (change-statement stmt rest)]
      [((kind ,k) ,rest ...)
       (set-statement-kind! stmt k)
       (change-statement stmt rest)]
      [((expr ,e) ,rest ...)
       (set-statement-expr! stmt (new-address e))
       (change-statement stmt rest)]
      [((stmts ,s) ,rest ...)
       (set-statement-statements! stmt (new-address s))
       (change-statement stmt rest)]
      [((scope ,sc) ,rest ...)
       (set-statement-assoc_scope! stmt (new-address sc))
       (change-statement stmt rest)]
      [((true ,s) ,rest ...)
       (set-statement-then_statement! stmt (new-address s))
       (change-statement stmt rest)]
      [((false ,s) ,rest ...)
       (set-statement-else_statement! stmt (new-address s))
       (change-statement stmt rest)]
      [((empty_else ,bool) ,rest ...)
       (set-statement-has_empty_else_clause! stmt bool)      
       (change-statement stmt rest)]
      [((init ,s) ,rest ...)
       (set-statement-initialization! stmt (new-address s))
       (change-statement stmt rest)]
      [((stmt ,s) ,rest ...)
       (set-statement-statement! stmt (new-address s))
       (change-statement stmt rest)]
      [((incr ,e) ,rest ...)
       (set-statement-increment! stmt (new-address e))
       (change-statement stmt rest)]
      [,else (error 'change-statement "invalid statement field:~s~n" else)])))

;
; Makes a new statement, attaches it to a specific memory region, and
; then modifies an arbitrary list of fields in the statement, using
; change-statement.
;
(define construct-statement
  (lambda args
    (if (or (null? args) (not (memory_region? (car args))))
	(error 'construct-statement "must receive a memory region and list of params")
	(let* ([mregion (car args)]
	       [new (new-statement mregion)])
	  (change-statement new (cdr args))
	  new))))

(define change-expression
  (lambda (expr args)
    (match args
      [() expr]
      [((next ,s) ,rest ...)
       (set-expression-next! expr (new-address s))
       (change-expression expr rest)]    
      [((kind ,k) ,rest ...)
       (set-expression-kind! expr k)
       (change-expression expr rest)]   
      [((opkind ,k) ,rest ...)
       (set-expression-operation_kind! expr k)
       (change-expression expr rest)]   
      [((routine ,r) ,rest ...)
       (set-expression-routine! expr (new-address r))
       (change-expression expr rest)]
      [((type ,t) ,rest ...)
       (set-expression-type! expr (new-address t))
       (change-expression expr rest)]   
      [((operands ,os) ,rest ...)
       (set-expression-operands! expr (new-address os))
       (change-expression expr rest)]   
      [((variable ,v) ,rest ...)
       (set-expression-variable! expr (new-address v))
       (change-expression expr rest)]   
      [((constant ,c) ,rest ...)
       (set-expression-constant! expr (new-address c))
       (change-expression expr rest)]   
      [((field ,f) ,rest ...)
       (set-expression-field! expr (new-address f))
       (change-expression expr rest)]  
      [((oddtype ,t) ,rest ...)
       (set-expression-odd_typing! expr t)
       (change-expression expr rest)]  
      [((eventype ,t) ,rest ...)
       (set-expression-even_typing! expr t)
       (change-expression expr rest)]  
      [((oddterms ,ts) ,rest ...)
       (set-expression-odd_terms! expr (new-address ts))
       (change-expression expr rest)]  
      [((eventerms ,ts) ,rest ...)
       (set-expression-even_terms! expr (new-address ts))
       (change-expression expr rest)]
      [((updated ,val) ,rest ...)
       (set-expression-updated! expr val)
       (change-expression expr rest)]       
      [((,invalid ,term) ,rest ...) 
       (error 'change-expression "invalid expression field ~s~n" invalid)]
      [,else
       (error 'change-expression "malformed arguments ~s~n" else)])))


(define construct-expression
  (lambda args
    (if (or (null? args) (not (memory_region? (car args))))
	(error 'construct-expression "must receive a memory region and list of params")
	(let* ([mregion (car args)]
	       [new (new-expression mregion)])
	  (change-expression new (cdr args))
	  new))))

(define get-highest-variable
  (lambda (mregion)
    (+ 1 (length (memory_region-variable_list mregion)))))

(define get-function-scope
  (lambda (mregion)
    (car (memory_region-scope_list mregion))))

(define is-static?
  (lambda (var mregion)
    (if (mempred var (get-list-of-records (a-r (scope-variables (get-function-scope mregion))) variable-next))
	#t
	#f)))

(define get-highest-param_type
  (lambda (mregion)
    (+ 1 (length (memory_region-param_type_list mregion)))))

(define clone-param_type
  (lambda (pt type mregion)
    (let ([newpt (make-param_type
		  (get-highest-param_type mregion)
		  (new-address (a-r (param_type-next pt)))
		  (new-address (a-r (param_type-type pt)))
		  type)])
      (set-memory_region-param_type_list!
       mregion
       (append
	(memory_region-param_type_list mregion)
	(list newpt)))
      newpt)))

(define new-variable
  (lambda (static? parameter? mregion)
    ;(printf "In mem region ~s, static: ~s~n" (memory_region-number mregion) static?)
    (if 
      (null? mregion)
      (error 'new-statement "new-statement must take a memory region as an argument~n"))
    (let* ([new (make-variable
		 (get-highest-variable mregion) 0 0 0 0 0 0 0 0 0
		 0                              0 0 0 0 0 0 0 0 0
		 0                              0 0 0 0 0 0 0 0
		 '() '() '() '())]
	   [scp (get-function-scope mregion)]
	   [first-var
	    (if static? 
		(new-address (a-r (scope-variables scp)))
		(new-address (a-r (scope-nonstatic_variables scp))))])
      
      (set-memory_region-variable_list! 
       mregion 
       (append (memory_region-variable_list mregion) (list new)))
      (if (not parameter?)
	  (begin
	    (set-variable-next! new first-var)
	    (let ([new-a (new-address new)])
	      ;(printf "~s~n" (map variable-name (get-list-of-records first-var variable-next)))
	      ;(printf "~s~n" (map variable-name (get-list-of-records (scope-nonstatic_variables scp) variable-next)))
	      (if static?
		  (set-scope-variables! scp new-a)
		  (set-scope-nonstatic_variables! scp new-a))
	      ;(printf "~s~n" (map variable-name (get-list-of-records (scope-nonstatic_variables scp) variable-next)))
	      )))
      new)))

(define change-variable
 (lambda (var args)
   (match args
     [() var]
     [((name ,n) ,rest ...)
      (set-variable-name! var n)
      (set-variable-unmangled_name! var n)
      (change-variable var rest)]
     [((next ,v) ,rest ...)
       (printf "here!")
       (set-variable-next! var (new-address v))
       (change-variable var rest)]    
     [((type ,t) ,rest ...)
      (set-variable-type! var (new-address t))
      (change-variable var rest)]
     [((storage_class ,sc) ,rest ...)
      (set-variable-storage_class! var sc)
      (change-variable var rest)]
     [((register ,v) ,rest ...)
      (set-variable-register! var (new-address v))
      (change-variable var rest)]
     [((odd_shadow ,v) ,rest ...)
      (set-variable-odd_shadow! var (new-address v))
      (change-variable var rest)]
     [((even_shadow ,v) ,rest ...)
      (set-variable-even_shadow! var (new-address v))
      (change-variable var rest)]
     [((is_parameter ,b) ,rest ...)
      (set-variable-is_parameter! var b)
      (change-variable var rest)])))

(define construct-variable
  (lambda args
    (if (or (null? args) (null? (cdr args)) (not (memory_region? (caddr args))))
	(begin 
	  ;(printf "args: ~s~n" args) 
	       (error 'construct-variable "must receive a memory region, a bool denoting staticness and a list of params"))
	(let* ([mregion (caddr args)]
	       [static? (car args)]
	       [param? (cadr args)]
	       [new (new-variable static? param? mregion)])
	  (change-variable new (cdddr args))
	  new))))

;
; Make a new if statement.
; 
(define new-if
  (lambda (test-expr true-stmt false-stmt next-stmt mregion)
    (construct-statement 
     mregion 
     `(next ,next-stmt)
     `(kind stmk_if)
     `(expr ,test-expr)
     `(true ,true-stmt)
     `(false ,false-stmt)
     `(empty_else ,(null? false-stmt)))))

(define string-exprs
  (lambda (exprs)
    (cond
     [(null? exprs) 0]
     [(null? (cdr exprs)) (change-expression (car exprs) `((next ())))]
     [else (change-expression (car exprs) `((next ,(cadr exprs)))) (string-exprs (cdr exprs))])))

;
; Make an enk_operation expression with given operands
;
(define new-op
  (lambda (kind next mregion . ops)
    (string-exprs ops)
    (construct-expression
     mregion
     `(kind enk_operation)
     `(opkind ,kind)
     `(operands ,(if (nil? ops) '() (car ops))))))

(define var->var_expr
  (lambda (var mregion)
    (construct-expression
     mregion
     `(kind enk_variable)
     `(variable ,var))))

(define var->var_addr_expr
  (lambda (var mregion)
    (construct-expression
     mregion
     `(kind enk_variable_address)
     `(variable ,var))))

(define var_addr_expr->var_expr
  (lambda (expr mregion)
    (construct-expression
     mregion
     `(kind enk_variable)
     `(variable ,(a-r (expression-variable expr))))))
  
(define copy-expr-replace-expr
  (lambda (expr old-expr new-expr mregion)
    (cond
     [(null? expr) '()]
     [(eq? expr old-expr) new-expr]
     [else
      (construct-expression
       mregion
       `(next ,(if (eq? (a-r (expression-next expr)) old-expr)
		   new-expr
		   (copy-expr-replace-expr (a-r (expression-next expr)) old-expr new-expr mregion)))
       `(kind ,(expression-kind expr))
       `(opkind ,(expression-operation_kind expr))
       `(variable ,(a-r (expression-variable expr)))
       `(routine ,(a-r (expression-routine expr)))
       `(constant ,(a-r (expression-constant expr)))
       `(field ,(a-r (expression-field expr)))
       `(operands ,(if (eq? (a-r (expression-operands expr)) old-expr)
		       new-expr
		       (copy-expr-replace-expr (a-r (expression-operands expr)) old-expr new-expr mregion))))])))
       
		   

(define copy-stmt-replace-expr
  (lambda (stmt old-expr new-expr mregion)
;    (printf "entering copy-stmt-replace-expr with stmt:~n~s~n" stmt)
    (cond
     [(null? stmt) '()]
     [else
      (construct-statement
       mregion
       `(next ())
       `(kind ,(statement-kind stmt))
       `(expr ,(if (eq? (a-r (statement-expr stmt)) old-expr)
		   new-expr
		   (copy-expr-replace-expr (a-r (statement-expr stmt)) old-expr new-expr mregion)))
       `(incr ,(if (eq? (a-r (statement-increment stmt)) old-expr)
		   new-expr
		   (copy-expr-replace-expr (a-r (statement-increment stmt)) old-expr new-expr mregion)))
       `(stmts ,(a-r (statement-statements stmt)))
       `(true ,(a-r (statement-then_statement stmt)))
       `(false ,(a-r (statement-else_statement stmt)))
       `(init ,(a-r (statement-initialization stmt)))
       `(stmt ,(statement-statement stmt)))])))
       
(define replace-statement
  (lambda (a b)
    (copy-statement a temp-statement)
    (copy-statement a b)
    (copy-statement temp-statement b)
    (delete-statement b)))

(define insert-before-statement
  (lambda (a b)
    (let ([a-ptr a]
	  [b-ptr b]
	  [a-next-ptr (if (address? (statement-next a))
			  (a-r (statement-next a))
			  ())])
      (copy-statement temp-statement a)
      (copy-statement a b)
      (copy-statement b temp-statement)
      (change-statement a `((next ,b-ptr)))
      (change-statement b `((next ,a-next-ptr))))))
		
(define insert-after-statement
  (lambda (a b)
    (change-statement a `((next ,(a-r (statement-next b)))))
    (change-statement b `((next ,a)))))
	  
(define stmt-with-question?
  (lambda (stmt)
    (let* ([exprs (stmt->exprs stmt)]
	   [first (first-match exprs (lambda (x) (equal? (expression-operation_kind x) 'eok_question)))])
      (if first #t #f))))
	   

(define fix-?-statement
  (lambda (stmt mregion)
    (let ([result (ormap-returns 
		   (lambda (x) (equal? (expression-operation_kind x) 'eok_question))
		   (stmt->exprs stmt))])
      (let* ([replacement-true-stmt 
	      (copy-stmt-replace-expr 
	       stmt 
	       result 
	       (a-r (expression-next (a-r (expression-operands result))))
	       mregion)]
	     [replacement-false-stmt 
	      (copy-stmt-replace-expr 
	       stmt 
	       result 
	       (a-r (expression-next (a-r (expression-next (a-r (expression-operands result))))))
	       mregion)] 
	     [replacement-if (new-if (a-r (expression-operands result))
				     replacement-true-stmt
				     replacement-false-stmt
				     (a-r (statement-next stmt))
				     mregion)])
; 	    (printf "result: ~s~n" result)
; 	    (printf "true: ~s~n" replacement-true-stmt)
; 	    (printf "false: ~s~n" replacement-false-stmt)
; 	    (printf "replacement-if: ~s~n" replacement-if)
; 	    (printf "replacement-if-false: ~s~n" (handle-statement (a-r (statement-else_statement replacement-if))))
	(replace-statement stmt replacement-if)))))

; (define remove=from-stmt
;   (lambda (stmt)
;     (if (or (null? stmt) (not (statement? stmt)) (not (equal? (statement-kind stmt) 'stmk_expr)))
; 	stmt
; 	(let* ([expr (strip-casts (a-r (statement-expr stmt)))]
; 	       [result (ormap-returns 
; 		   (lambda (x) (equal? (expression-operation_kind x) 'eok_assign))
; 		   (reverse (stmt->exprs stmt)))])
; 	  (if (or (equal? result #f) (equal? result expr))
; 	      #t
	      


(define opie-id-count 0)
 
;
; Return a fresh number as a string
;
(define fresh-opie-id
  (lambda ()
    (set! opie-id-count (+ 1 opie-id-count))
    (number->string opie-id-count)))
;
; Create a register for a variable.  It will be named: 
;
;   Opie415_reg_<fresh id>_<var name>
;
; Where fresh id is generated by fresh-opie-id and variable
; name is the name of the variable requiring a register.
;
(define create-reg-var
  (lambda (var mregion)
    (let ([register (construct-variable 
		     (is-static? var mregion)
		     (variable-is_parameter var)
		     mregion 		     
		     `(type ,(a-r (variable-type var)))
		     `(name ,(string->symbol (string-append "Opie415_reg_" (fresh-opie-id) "_" (symbol->string (variable-name var))))))])
      (change-variable var `((odd_shadow ,register)))))) 

;
; Create an odd shadow of a variable with name:
;
; Opie415_odd_<fresh id>_<variable name>
;
(define create-odd-shadow
  (lambda (var mregion)
    (let ([shadow (construct-variable
		   (is-static? var mregion)
		   (variable-is_parameter var)		   
		   mregion
		   `(type ,(a-r (variable-type var))) 
		   `(name 		     
		     ,(string->symbol 
		       (string-append 
			"Opie415_odd_" 
			(fresh-opie-id) 
			"_" 
			(symbol->string (variable-name var)))))
		   `(storage_class ,(variable-storage_class var))
		   `(is_parameter ,(variable-is_parameter var))
		   ;`(next ,(a-r (variable-next var)))
		   )])
      (change-variable var `((odd_shadow ,shadow)))
'      (change-variable var `((next ,shadow)))
      (printf "~s ~s~n" (variable-name var) (variable-is_parameter var))
      (if (variable-is_parameter var)
	  (let ([cpt (clone-param_type (a-r (variable-assoc_param_type var)) 'simple_odd mregion)])
	    (printf "cloned: ~s~n" cpt)
            (change-variable shadow `((next ,(a-r (variable-next var)))))
	    (change-variable var `((next ,shadow)))
	    (set-param_type-next! (a-r (variable-assoc_param_type var)) (new-address cpt)))))))
 
;
; Create an even shadow of a variable with name:
;
; Opie415_even_<fresh id>_<variable name>
;
(define create-even-shadow
  (lambda (var mregion)
    (let ([shadow (construct-variable
		   (is-static? var mregion)
		   (variable-is_parameter var)		   
		   mregion
		   `(type ,(a-r (variable-type var))) 
		   `(name 		     
		     ,(string->symbol 
		       (string-append 
			"Opie415_even_" 
			(fresh-opie-id) 
			"_" 
			(symbol->string (variable-name var)))))
		   `(storage_class ,(variable-storage_class var))
		   `(is_parameter ,(variable-is_parameter var))
		   ;`(next ,(a-r (variable-next var)))
		   )])
      (change-variable var `((even_shadow ,shadow)))
;      (change-variable var `((next ,shadow)))
      (printf "~s ~s~n" (variable-name var) (variable-is_parameter var))
      (if (variable-is_parameter var)
	  (let ([cpt (clone-param_type (a-r (variable-assoc_param_type var)) 'simple_odd mregion)])
	    (printf "cloned: ~s~n" cpt)
	    (change-variable shadow `((next ,(a-r (variable-next var)))))
	    (change-variable var `((next ,shadow)))
	    (set-param_type-next! (a-r (variable-assoc_param_type var)) (new-address cpt)))))))

;
; Create an even shadow of a variable with name:
;
; Opie415_even_<fresh id>_<variable name>
;
; (define create-even-shadow
;   (lambda (var mregion)
;     (let ([shadow (construct-variable
; 		   (is-static? var mregion)	
; 		   (variable-is_parameter var)	   
; 		   mregion
; 		   `(type ,(a-r (variable-type var)))
; 		   `(name ,(string->symbol 
; 			    (string-append 
; 			     "Opie415_even_" 
; 			     (fresh-opie-id) 
; 			     "_" 
; 			     (symbol->string (variable-name var)))))
; 		   `(storage_class ,(variable-storage_class var))
; 		   `(is_parameter ,(variable-is_parameter var))
; ;		   `(next ,(a-r (variable-next var)))
; 		   )])
;       (printf "~s ~s~n" (variable-name var) (variable-is_parameter var))
;       (change-variable var `((even_shadow ,shadow)))
; ;      (change-variable var `((next ,shadow)))
;       (if (variable-is_parameter var)
; 	  (let ([cpt (clone-param_type (a-r (variable-assoc_param_type var)) 'simple_even mregion)])
; 	    (change-variable var `((next ,shadow)))
; 	    (set-param_type-next! (a-r (variable-assoc_param_type var)) (new-address cpt)))))))

(define shadow-of
  (lambda (var)
    (if (address? var)
	(error 'shadow-of "shadow-of expects an actual variable, not an address"))
    (if (nil? (variable-odd_shadow var))
	(if (nil? (variable-even_shadow var))
	    (error 'shadow-of "variable ~s has no shadow" var)
	    (a-r (variable-even_shadow var)))
	(a-r (variable-odd_shadow var)))))

; (define make-shadow-assignment   (lambda (assignment-type var rhs even? mregion)
;     (let ([rhs (strip-casts rhs)])
;       (cond
;        [(equal? (expression-kind rhs) 'enk_constant)
; 	(let* ([const (a-r (expression-constant rhs))]
; 	       [intval (integer_constant_tail-integer_value (constant-constant_tail const))])
; 	  (if (or (equal? intval `~) (null? intval))
; 	      (error 'make-shadow-assignment "rhs constant has no integer value: ~s" const)
; 	      (let ([new-rhs (if even?
; 				 (new-int-const (dilate-int intval) mregion) 
; 				 (new-int-const (string->symbol 
; 							  (string-append 
; 							   "2*"
; 							   (symbol->string (dilate-int intval))
; 							   ))
; 							 mregion))])
; 		(let ([new-lhs (var->var_expr (shadow-of var) mregion)])
; 		    (change-expression new-lhs `((next ,new-rhs)))
; 		    (printf "new-rhs: ~s~n" new-rhs)
; 		    (new-op assignment-type `() mregion new-lhs new-rhs)))))]
;        [else (error 'make-shadow-assignment "couldn't match rhs of kind: ~s" (expression-kind rhs))]))))						
(define make-shadow-assignment
  (lambda (assignment-type var rhs even? mregion)
    (let* ([rhs (strip-casts rhs)]	   
	   [new-rhs (shadow-expr rhs even? mregion)]
	   [new-lhs (var->var_expr (shadow-of var) mregion)]
	   [new-assignment-type (case assignment-type
				  [(eok_iassign) 'eok_iassign]
				  [(eok_iadd_assign) (if even? 'eok_even_add_assign 'eok_odd_add_assign)]
				  [else (error 'make-shadow-assignment "unmatched assignment type ~s" assignment-type)])])
      (change-expression new-lhs `((next ,new-rhs)))
      (new-op new-assignment-type `() mregion new-lhs new-rhs))))
     				    
(define assignment-kind?
  (lambda (opkind)
    (if (or (equal? opkind 'eok_iassign)
	    (equal? opkind 'eok_iadd_assign))
	#t
	#f)))

(define shadow-assignment
  (lambda (expr mregion)
    ;(printf "~nshadow-assignment with: ~n~s~n~s~n~s~n~n" expr (expression-next expr) (handle-expression expr))
    (cond
     [(not (equal? (expression-kind expr) 'enk_operation)) (change-expression expr `((updated ,#t)))]
     [else (let ([opkind (expression-operation_kind expr)])
	     (if (assignment-kind? opkind)
		 (let* ([op1 (a-r (expression-operands expr))]
			[op2 (a-r (expression-next op1))]
			[var (a-r (expression-variable op1))])
		   (if (or (nil? var) (nil? op1) (nil? op2))
		       (error 'shadow-assignment "lhs has a nil variable or operand: ~s~n~s~n~s~n" op1 op2 var)
		       (if (nil? (variable-even_shadow var))
			   (if (nil? (variable-odd_shadow var))
			       (begin
				 (printf "--- tried to shadow a non-shadowed variable assignment ---~n")
				 (printf "--- assignment will be unchanged ---~n")
				 (change-expression expr `((updated ,#t))))
			       (let* ([new-assign (make-shadow-assignment opkind var op2 #f mregion)]
				      [comma-expr (new-op 'eok_comma `() mregion)])
				 (swap-expressions comma-expr expr)
				 (change-expression new-assign `((next ,comma-expr) (updated ,#t)))
				 (change-expression expr `((operands ,new-assign)))
				 (change-expression comma-expr `((updated ,#t)))
				 ))
			   (let* ([new-assign (make-shadow-assignment opkind var op2 #t mregion)]
				  [comma-expr (new-op 'eok_comma `() mregion)])
			     (swap-expressions comma-expr expr)
			     (change-expression new-assign `((next ,comma-expr) (updated ,#t)))
			     (change-expression expr `((operands ,new-assign)))
			     (change-expression comma-expr `((updated ,#t))))))
		   (printf "shadow assignment: returning ~s~n" (handle-expression expr)))		 
		 (change-expression expr `((updated ,#t)))))])))
			  
					       
			    
		    
		    


(define non-updated-increment
  (lambda (expr)
    (if (expression-updated expr)
	#f
	(let ([opkind (expression-operation_kind expr)])
	  (if (or (equal? opkind 'eok_ipost_incr)
		  (equal? opkind 'eok_ipre_incr)
		  (equal? opkind 'eok_ipost_decr)
		  (equal? opkind 'eok_ipre_decr))
	      #t
	      #f)))))
	    
(define non-updated-shadow-assignment
  (lambda (expr)
    (if (expression-updated expr)
	#f
	(let ([opkind (expression-operation_kind expr)])
	  (if (assignment-kind? opkind)
	      (let* ([op1 (a-r (expression-operands expr))]
		     [op1kind (expression-kind op1)])
		;(printf "op1kind ~s~n" op1kind)
		(if (or (equal? op1kind 'enk_variable) (equal? op1kind 'enk_variable_address))
		    (shadowed? (a-r (expression-variable op1)))
		    (let ([op1opkind (expression-operation_kind op1)])
		      ;(printf "op1opkind ~s~n" op1opkind)
		      ;(printf "op11 ~s~n" (handle-expression (a-r (expression-operands op1))))
		      ;(printf "op12 ~s~n" (handle-expression (a-r (expression-next (a-r (expression-operands op1))))))
		      #f)))
	      #f)))))

(define shadow-increment
  (lambda (expr mregion)
    ;(printf "entered shadow-increment with ~s~n" (expression-operation_kind expr))
    (let ([opkind (expression-operation_kind expr)])
      (cond
       [(nil? expr) #f]
       [(or (equal? opkind 'eok_ipost_incr)
	    (equal? opkind 'eok_ipre_incr)
	    (equal? opkind 'eok_ipost_decr)
	    (equal? opkind 'eok_ipre_decr))
	(let* ([operand (strip-casts (a-r (expression-operands expr)))]
	       [var (if (or (equal? (expression-kind operand) 'enk_variable) 
			  (equal? (expression-kind operand) 'enk_variable_address))
			(a-r (expression-variable operand))
			(error 'shadow-increment "increment tried to shadow operand of kind ~s~n" (expression-kind operand)))])
	  (set-expression-updated! expr #t)
	  (if (even-shadowed? var)
	      (let* ([newopkind (if (or (equal? opkind 'eok_ipost_incr) (equal? opkind 'eok_ipre_incr))
				    'eok_even_inc
				    'eok_even_decr)]
		     [newop (var->var_expr (shadow-of var) mregion)]
		     [newexpr1 (new-op newopkind `() mregion newop)]
		     [address-temp (a-r (expression-next expr))])
		(change-expression expr `((next ,newexpr1)))
		;(printf "  ~s~n" (a-r (expression-next expr)))
		(let ([newexpr2 (new-op 'eok_comma address-temp mregion)])
		  (change-expression newexpr2 `((operands ,newexpr2)))
		  (swap-expressions expr newexpr2)
		  ;(printf "  ~s~n" (a-r (expression-next newexpr2)))
		  ))
	      (if (odd-shadowed? var)
		  (let* ([newopkind (if (or (equal? opkind 'eok_ipost_incr) (equal? opkind 'eok_ipre_incr))
					'eok_odd_inc
					'eok_odd_decr)]
			 [newop (var->var_expr (shadow-of var) mregion)]
			 [newexpr1 (new-op newopkind `() mregion newop)]
			 [address-temp (expression-next expr)])
		    (change-expression expr `((next ,newexpr1)))
		    ;(printf "  ~s~n" (a-r (expression-next expr)))
		    (let ([newexpr2 (new-op 'eok_comma address-temp mregion)])
		      (change-expression newexpr2 `((operands ,newexpr2)))
		      (swap-expressions expr newexpr2)
					;(printf "  ~s~n" (a-r (expression-next newexpr2)))
		  )))))]))))
  
; (define shadow-increment
;   (lambda (stmt mregion)
;     ;(printf "entered shadow-increment with ~s~n~s~n~s~n" stmt (a-r (statement-expr stmt)) (statement-kind stmt))
;     (cond
;      [(nil? stmt) #f]
;      [(not (equal? (statement-kind stmt) 'stmk_expr)) #f]
;      [(nil? (a-r (statement-expr stmt))) #f]
;      [(not (equal? (expression-kind (a-r (statement-expr stmt))) 'enk_operation)) #f]
;      [else
;       (let* ([expr (a-r (statement-expr stmt)) ]
;	       [opkind (expression-operation_kind expr)])
; 	(if (or (equal? opkind 'eok_ipost_incr)
; 		(equal? opkind 'eok_ipre_incr)
; 		(equal? opkind 'eok_ipost_decr)
; 		(equal? opkind 'eok_ipre_decr))
; 	    (let* ([operand (strip-casts (a-r (expression-operands expr)))]
; 		   [var (if (or (equal? (expression-kind operand) 'enk_variable) 
; 				(equal? (expression-kind operand) 'enk_variable_address))
; 			    (a-r (expression-variable operand))
; 			    (error 'shadow-increment "increment tried to shadow operand of kind ~s~n" (expression-kind operand)))])
; 	      (if (even-shadowed? var)
; 		  (let* ([newopkind (if (or (equal? opkind 'eok_ipost_incr) (equal? opkind 'eok_ipre_incr))
; 					'eok_even_inc
; 					'eok_even_decr)]
; 			 [newop (var->var_expr (shadow-of var) mregion)]
; 			 [newexpr (new-op newopkind '() mregion newop)]
; 			 [newstmt (construct-statement mregion `(kind stmk_expr) `(expr ,newexpr))])
; 		    (insert-after-statement newstmt stmt)
; 		    #t)
; 		  (if (odd-shadowed? var)
; 		      (let* ([newopkind (if (or (equal? opkind 'eok_ipost_incr) (equal? opkind 'eok_ipre_incr))
; 					    'eok_odd_inc
; 					    'eok_odd_decr)]
; 			  [newop (var->var_expr (shadow-of var) mregion)]
; 			  [newexpr (new-op newopkind '() mregion newop)]
; 			  [newstmt (construct-statement mregion `(kind stmk_expr) `(expr ,newexpr))])
; 			(insert-after-statement newstmt stmt)
; 			#t)		   
; 		      #f)))
; 	    #f))])))

	     
; (define shadow-increment
;   (lambda (stmt mregion)
;     (let* ([shadow (shadow-of 
; 		    (a-r 
; 		     (expression-variable 
; 		      (strip-casts 
; 		       (a-r 
; 			(expression-operands 
; 			 (a-r 
; 			  (statement-expr stmt))))))))]
; 	   [new (construct-statement
; 		mregion
; 		`(next ,(a-r (statement-next stmt)))
; 		`(kind stmk_expr)
; 		`(expr ,(construct-expression
; 			 mregion
; 			 `(kind ,(expression-kind (a-r (statement-expr stmt))))
; 			 `(opkind ,(expression-operation_kind (a-r (statement-expr stmt))))
; 			 `(operands ,(construct-expression
; 				      mregion
; 				      `(kind enk_variable)
; 				      `(variable ,shadow))))))])
;       (insert-before-statement stmt new))))
	       
(define shadowed?
  (lambda (var)
    (or (even-shadowed? var) (odd-shadowed? var))))

(define even-shadowed?
  (lambda (var)
    (not (nil? (variable-even_shadow var)))))

(define odd-shadowed?
  (lambda (var)
    (not (nil? (variable-odd_shadow var)))))

(define shadow-expr
  (lambda (expr even? mregion)
    (printf "entered shadow-expr with ~s~n" (handle-expression expr))
    (cond
     [(nil? expr) (error 'shadow-expr "nil expr~n")]
     [(equal? (expression-kind expr) 'enk_operation)
      (match (expression-operation_kind expr)
	[~  (error 'shadow-expr "unitialized op kind~n")]
	[() (error 'shadow-expr "unitialized op kind~n")]
	[eok_iadd 
	 (new-op 
	  (if even? 'eok_even_add 'eok_odd_add) 
	  '() 
	  mregion 
	  (shadow-expr (a-r (expression-operands expr)) even? mregion)
	  (shadow-expr (a-r (expression-next (a-r (expression-operands expr)))) even? mregion))]
	[eok_isubtract
	 (new-op 
	  (if even? 'eok_even_sub 'eok_odd_sub) 
	  '() 
	  mregion 
	  (shadow-expr (a-r (expression-operands expr)) even? mregion)
	  (shadow-expr (a-r (expression-next (a-r (expression-operands expr)))) even? mregion))]
	[eok_ipost_incr
	 (new-op
	  (if even? 'eok_even_inc 'eok_odd_inc) 
	  '() 
	  mregion 
	  (shadow-expr (a-r (expression-operands expr)) even? mregion))]
	[eok_ipre_incr
	 (new-op
	  (if even? 'eok_even_inc 'eok_odd_inc) 
	  '() 
	  mregion 
	  (shadow-expr (a-r (expression-operands expr)) even? mregion))]
	[,else (error 'shadow-expr "unable to handle operation of kind ~s~n" else)])]
     [(equal? (expression-kind expr) 'enk_variable)
      (let ([var (a-r (expression-variable expr))])
	;(printf "shadowing a var ~s~n" var)
	(if even?
	    (if (even-shadowed? var)
		(var->var_expr (shadow-of var) mregion)
		(if (odd-shadowed? var)
		    (new-op 'eok_odd_to_even '() mregion (var->var_expr (shadow-of var) mregion))
		    (begin
		      (create-even-shadow var mregion)
		      (var->var_expr (shadow-of var) mregion))))
		    ;(error 'shadow-expr "missed a secondary shadow~n")))
	    (if (odd-shadowed? var)
		(var->var_expr (shadow-of var) mregion)
		(if (even-shadowed? var)
		    (new-op 'eok_even_to_odd '() mregion (var->var_expr (shadow-of var) mregion))
		    (begin
		      (create-odd-shadow var mregion)
		      (var->var_expr (shadow-of var) mregion))))))]
		    ;(error 'shadow-expr "missed a secondary shadow~n")))))]
     [(equal? (expression-kind expr) 'enk_constant)
      (let* ([const (a-r (expression-constant expr))]
	     [intval (integer_constant_tail-integer_value (constant-constant_tail const))]
	     [new-const (if even?
			    (new-int-const (dilate-int intval) mregion) 
			    (new-int-const (string->symbol 
					    (string-append 
					     "2*"
					     (symbol->string (dilate-int intval))
					     ))
					   mregion))])
	new-const)]	    
     [else (error 'shadow-expr "unable to handle expression of kind ~s~n" (expression-kind expr))])))

(define shadow-iassign
  (lambda (stmt mregion)
    (cond
     [(nil? stmt) #f]
     [(not (and (statement? stmt)
		(equal? (statement-kind stmt) 'stmk_expr)
		(not (nil? (statement-expr stmt)))
		(equal? (expression-kind (a-r (statement-expression stmt))) 'enk_operation)
		(equal? (expression-operation_kind (a-r (statement-expression stmt))) 'eok_iassign)))
      #f]
     [else 
      (let* ([assign-expr (a-r (statement-expression stmt))]
	     [assign-left (a-r (expression-operands assign-expr))]
	     [assign-right (a-r (expression-next assign-left))]
	     [new (new-iassign (shadow-of (a-r (expression-variable assign-left))) (shadow-expr assign-left mregion))])
      (insert-before-stmt stmt new))])))
	
(define conditional-even-shadow
  (lambda (var mregion)
    ;(printf "c-e-s: ~s~n~s ~s~n" var (even-shadowed? var) (odd-shadowed? var))
    (if (not (even-shadowed? var))
	(create-even-shadow var mregion))))

(define conditional-odd-shadow
  (lambda (var mregion)
    ;(printf "c-o-s: ~s~n~s ~s~n" var (even-shadowed? var) (odd-shadowed? var))
    (if (not (or (even-shadowed? var) (odd-shadowed? var)))
	(create-odd-shadow var mregion))))

(define shadow-matrix-vars-even
  (lambda (matrix mregion)
    (map (lambda (var) (conditional-even-shadow var mregion)) (one_matrix-even_vars matrix))))

(define shadow-matrix-vars-odd
  (lambda (matrix mregion)
    (map (lambda (var) (conditional-odd-shadow var mregion)) (one_matrix-odd_vars matrix))))

; (define shadow-rhs
;   (lambda (stmt mregion)
;     (let ([expr (a-r (statement-expr stmt))])
;       (if (and (equal? (statement-kind stmt) 'stmk_expr)
; 	       (equal? (expression-kind expr) 'enk_operation)
; 	       (equal? (expression-operation_kind expr) 'eok_iassign))
; 	  (let* ([lhs (a-r (expression-operands expr))]
; 		 [lhs-var (a-r (expression-variable lhs))]
; 		 [rhs (get-next lhs)]
; 		 [vars (expr->vars rhs)])
; 	    (if (even-shadowed? lhs-var)
; 		(shadow-list-of-vars #t vars)
; 		(if (odd-shadowed? lhs-var)
; 		    (map (lambda (var) (if (not (or (odd-shadowed) (even-shadowed? var)))
; 					   (begin (create-odd-shadow var mregion) (shadow-rhss mregion)))))
; 		    vars))))))))


(define shadow-rhs-vars
  (lambda (vars even? mregion)
    (cond
     [(null? vars) #f]
     [(or (even-shadowed? (car vars)) (odd-shadowed? (car vars))) (shadow-rhs-vars (cdr vars) even? mregion)]
     [else (if even?

	       (create-even-shadow (car var) mregion)
	       (create-odd-shadow (car var) mregion))
	   (or (shadow-rhs-vars (cdr vars)) #t)])))
     

(define shadow-rhs-statement
  (lambda (stmt mregion)
    (let ([expr (a-r (statement-expr stmt))])
      (if (and (equal? (statement-kind stmt) 'stmk_expr)
	       (equal? (expression-kind expr) 'enk_operation)
	       (equal? (expression-operation_kind expr) 'eok_iassign))
	  (let* ([lhs (a-r (expression-operands expr))]
		 [lhs-var (a-r (expression-variable lhs))]
		 [rhs (get-next lhs)]
		 [vars (expr->vars rhs)])
	    (if (even-shadowed? lhs-var)
		(shadow-rhs-vars vars #t mregion)
		(if (odd-shadowed? lhs-var)
		    (shadow-rhs-vars vars #f mregion)
		    #f)))))))

(define list-or
  (lambda (ls)
    (cond
     [(null? ls) #f]
     [(car ls) #t]
     [else (list-or (cdr ls))])))

(define shadow-rhs-mregion
  (lambda (mregion)
    (let* ([stmts (memory_region-statement_list mregion)]
	   [result (list-or (map (lambda (stmt) (shadow-rhs-statement stmt mregion)) stmts))])
      (if result
	  (shadow-rhs-mregion mregion)))))
 

(define shadow-vars
  (lambda (mregion)
    (map (lambda (matrix) (shadow-matrix-vars-even matrix mregion)) (memory_region-matrix_info mregion))
    (map (lambda (matrix) (shadow-matrix-vars-odd matrix mregion)) (memory_region-matrix_info mregion))
    ;(shadow-rhss mregion)
))

(define shadow-assignments
  (lambda (mregion)
    (let ([get-expr-list (lambda () (memory_region-expression_list mregion))])
;    (map (lambda (stmt) (shadow-iassign stmt mregion)) (memory_region-statement_list mregion))
      ;(printf "shadowing = assignments~n")
      (map (lambda (expr) (set-expression-updated! expr #f)) (memory_region-expression_list mregion))
      (execute-while-exists get-expr-list non-updated-shadow-assignment shadow-assignment mregion)
      ;(printf "shadoing increments~n")
      (map (lambda (expr) (set-expression-updated! expr #f)) (memory_region-expression_list mregion))
      (execute-while-exists get-expr-list non-updated-increment shadow-increment mregion))))

(define new-int-const
  (lambda (num mregion)
    (construct-expression 
     mregion
     `(kind enk_constant)
     `(constant ,(make-constant
		  (+ 1 (length (memory_region-constant_list mregion)))
		  '() '() '() '() '() '() '() '() '() '() '() '() '() '() '() '() '() '() '() 'ck_integer 
		  (make-integer_constant_tail num))))))

(define odd_term->expr
  (lambda (term mregion)
    ;(printf "odd_term->expr on ~s~n" term)
    (cond 
     [(or (null? term) (not (list? term))) (error 'odd_term->expr "nonlist or null argument")]
     [(null? (cdr term)) (shadow-expr (car term) #f mregion)]
     [else (new-op 'eok_imultiply
		   '()
		   mregion
		   (car term)
		   (odd_term->expr (cdr term) mregion))])))

(define odd_terms_list->expr
  (lambda (terms mregion)
    ;(printf "odd_terms_list->expr on ~s~n" terms)
    (cond
     [(null? terms) '()]
     [(null? (cdr terms)) 
      (cond
       [(member? 'neg_one_const (car terms)) 
	(if (null? (cdar terms))
	    (new-int-const -1 mregion)
	    (new-op 'eok_imultiply
		    '()
		    mregion
		    (new-int-const -1 mregion)
		    (odd_term->expr (remove 'neg_one_const (car terms)) mregion)))]
       [(member? 'one_const (car terms))
	(if (null? (cdar terms))
	    (new-int-const 1 mregion)
	    (odd_term->expr (remove 'one_const (car terms)) mregion))]
       [else (odd_term->expr (car terms) mregion)])]
      [(member? 'neg_one_const (car terms))
       (new-op 'eok_odd_sub
	       '()
	       mregion
	       (odd_term->expr (remove 'neg_one_const (car terms)) mregion)
	       (odd_terms_list->expr (cdr terms) mregion))]
      [(member? 'one_const (car terms))
       (new-op 'eok_odd_add
	       '()
	       mregion
	       (remove 'one_const (odd_term->expr (car terms) mregion))
	       (odd_terms_list->expr (cdr terms) mregion))]
      [(new-op 'eok_odd_add
	       '()
	       mregion
	       (odd_term->expr (car terms) mregion)
	       (odd_terms_list->expr (cdr terms) mregion))])))

(define even_term->expr
  (lambda (term mregion)
    ;(printf "even_term->expr on ~s~n" term)
    (cond 
     [(or (null? term) (not (list? term))) (error 'odd_term->expr "nonlist or null argument")]
     [(equal? (car term) 'one_const) (new-int-const 1 mregion)]
     [(null? (cdr term)) (shadow-expr (car term) #t mregion)]
     [else (error 'even_term->expr "can't handle multiplication yet")])))

(define even_terms_list->expr
  (lambda (terms mregion)
    ;(printf "even_terms_list->expr on ~s~n" terms)
    (cond
     [(null? terms) '()]
     [(null? (cdr terms)) 
      (cond
       [(member? 'neg_one_const (car terms)) 
	(if (null? (cdar terms))
	    (new-int-const -2 mregion)
	    (new-op 'eok_imultiply
		    '()
		    mregion
		    (new-int-const -1 mregion)
		    (even_term->expr (remove 'neg_one_const (car terms)) mregion)))]
       [(member? 'one_const (car terms))
	(if (null? (cdar terms))
	    (new-int-const 1 mregion)
	    (even_term->expr (remove 'one_const (car terms)) mregion))]
       [else (even_term->expr (car terms) mregion)])]
      [(member? 'neg_one_const (car terms)) 
       (new-op 'eok_even_sub
	       '()
	       mregion
	       (even_term->expr (remove 'neg_one_const (car terms)) mregion)
	       (even_terms_list->expr (cdr terms) mregion))]
      [(member? 'one_const (car terms))
       (new-op 'eok_even_add
	       '()
	       mregion
	       (even_term->expr (remove 'one_const (car terms)) mregion)
	       (even_terms_list->expr (cdr terms) mregion))]
      [(new-op 'eok_even_add
	       '()
	       mregion
	       (even_term->expr (car terms) mregion)
	       (even_terms_list->expr (cdr terms) mregion))])))

(define shadow-matrix-deref
  (lambda (expr mregion)
    (let* ([o-terms (expression-odd_terms expr)]
	   [e-terms (expression-even_terms expr)]
	   [o-expr (odd_terms_list->expr o-terms mregion)]
	   [e-expr (even_terms_list->expr e-terms mregion)]
	   [new-deref (if (nil? o-expr)
			  (if (nil? e-expr)
			      (error 'shadow-matrix-deref "nil even and odd results")
			      e-expr)
			  (if (nil? e-expr)
			      o-expr
			      (new-op 'eok_iadd '() mregion o-expr e-expr)))])
      (swap-expressions expr new-deref))))

(define shadow-matrix
  (lambda (matrix mregion)
    (map (lambda (expr) (shadow-matrix-deref expr mregion)) (one_matrix-deref_exprs matrix))))

(define shadow-matrices
  (lambda (mregion)
    (map (lambda (matrix) (shadow-matrix matrix mregion)) (memory_region-matrix_info mregion))))

(define shadow-all
  (lambda (mregion)
    ;(printf "shading variables~n")
    (shadow-vars mregion)
    ;(printf "shadowing assignments~n")
    (shadow-assignments mregion)
    ;(printf "shadowing matrix dereferences~n")
    (shadow-matrices mregion)))

(define remove-assignment
  (lambda (stmt mregion)
    1))

(define fix-mregion
  (lambda (mregion)
    (let ([get-stmt-list (lambda () (memory_region-statement_list mregion))])
      ;(execute-while-exists get-stmt-list for-with-question-initialization? fix-for-initialization mregion)
      ;(execute-while-exists get-stmt-list for-with-question-test? fix-for-test mregion)
      ;(execute-while-exists get-stmt-list for-with-question-increment? fix-for-increment mregion)
      ;(map (lambda (stmt) (remove-assignment stmt mregion)) (memory_region-statement_list mregion))
      (execute-while-exists get-stmt-list stmt-with-question? fix-?-statement mregion)
      )))

(define parse-mregions
  (lambda (ss)
    ;(printf "parsing pragma information~n")
    ;(map set-matrix-info-strides (scil_structure-fun_mem_region_list ss))
    ;(printf "locating matrix dereferences~n")
    (map set-matrix-info-derefs (scil_structure-fun_mem_region_list ss))
    ;(printf "parsing matrix dereferences~n")
    (map parse-matrix-derefs (scil_structure-fun_mem_region_list ss))
    ;(printf "determining shadoing of matrix dereferences~n")
    (map set-matrix-info-oddeven (scil_structure-fun_mem_region_list ss))))

(define fix-mregions
  (lambda (ss)
    ;(printf "making some changes to C grammer~n")
    (map fix-mregion (scil_structure-fun_mem_region_list ss))))

(define shadow-mregions
  (lambda (ss)
    (map shadow-all (scil_structure-fun_mem_region_list ss))))



(define get-param-typing
  (lambda (pt)
    (list
      (string->symbol (flatten-and-string-append (handle-type (a-r (param_type-type pt)))))
      (param_type-dilated_type pt))))

(define number-param-typing
  (lambda (pts n)
    (match pts
      [() '()]
      [((,t untyped) . ,rest) (cons (cons (+ 1 n) `(,t untyped)) (number-param-typing rest (+ 1 n)))]
      [((,t ,nt) . ,rest) (cons (cons n `(,t ,nt)) (number-param-typing rest n))])))

(define type-func
  (lambda (routine)
    (if (not (routine-defined routine))
	#f
	(let* ([type (a-r (routine-type routine))]
	       [pts (routine_type_tail-param_type_list (type-type_tail type))]
	       [typing (map get-param-typing (get-list-of-records pts param_type-next))]
	       [numbered (number-param-typing typing -1)])
	  (set-routine-param_typing! routine numbered)))))

(define type-mregions
  (lambda (ss)
    (map type-func (memory_region-routine_list (scil_structure-file_scope ss)))))

(define morton-order-ss
  (lambda (ss)
    (parse-mregions ss)
    (fix-mregions ss)
    (shadow-mregions ss)
    (type-mregions ss) 
    ss))


; Take a triple and churn out a pragma structure
;   first element is the pragma type
;   second element is the target name
;   third element is the pragma data
;   fourth element is the target function
(define make-pragma-struct
  (lambda (pr)
    (make-pragma-record
      (car pr)
      (cadr pr)
      (caddr pr)
      (cadddr pr))))

; Read pragma data from a file
(define read-pragma-data
  (lambda (f)
     (with-input-from-file f
       (lambda ()
    	 (let ([pragma-list (read)])
    	   pragma-list)))))



; update all function calls in scil by comparinjg all calls
; to source functions in scils.  Also look for calls that involve
; morton matrices but have no matching sources -- error for these.
(define update-function-calls
  (lambda (scils)
    (lambda (scil)
;      (match-calls scil scils)
;      (update-calls scil)
;      (update-shadows scil)
      scil)))

(define zip
  (lambda (xs ys)
    (if (or (null? xs) (null? ys))
        '()
        (cons (cons (car xs) (car ys))
              (zip (cdr xs) (cdr ys))))))
              
(define add-extension              
  (lambda (ext files)
    (cond
      [(null? files) '()]
      [else (cons (string-append (car files) ext) (add-extension ext (cdr files)))])))

(define find-mem-region
  (lambda (n mlist)
    (cond
     [(null? mlist) (error 'find-mem-region "memory region ~s not found ~n" n)]
     [(equal? (memory_region-number (car mlist)) n) (car mlist)]
     [else (find-mem-region n (cdr mlist))])))

(define find-func-helper
  (lambda (rlist name scil)
    ;(printf "here with ~s~n" (if (not (null? rlist)) (routine-unmangled_name (car rlist)) 'nil))
    (cond
      [(null? rlist) #f]
      [(equal? (routine-name (car rlist)) name)
       (begin	
	 ;(printf "~s~n" (scil_structure-fun_mem_region_list scil))
	 (find-mem-region (routine-assoc_scope (car rlist)) (scil_structure-fun_mem_region_list scil)))]
;	 (list-ref (scil_structure-fun_mem_region_list scil) (- (routine-assoc_scope (car rlist)) 2)))]
      [else (find-func-helper (cdr rlist) name scil)])))

(define find-func
  (lambda (scil name)
    (if (scil_structure? scil)
	(find-func-helper (memory_region-routine_list (scil_structure-file_scope scil)) name scil)
	#f)))

(define find-var-helper
  (lambda (vlist name)
    (cond
     [(not (list? vlist)) #f]
     [(null? vlist) #f]
     [(equal? (variable-name (car vlist)) name)
      (car vlist)]
     [else (find-var-helper (cdr vlist) name)])))

(define find-var
  (lambda (mregion name)
    (if (memory_region? mregion)
	(find-var-helper (memory_region-variable_list mregion) name)
	#f)))
      
(define add-matrix-pragma
  (lambda (scil name stride func)
    (let* ([mregion (find-func scil func)]
	   [mvar (find-var mregion name)]
	   [svar (find-var mregion stride)])
      ;(printf "~s ~s ~s~n" name stride func)
      (cond
       [(not mregion)
	(error 'add-matrix-pragma "function not found ~s~n" func)]
       [(not mvar)
	(error 'add-matrix-pragma "matrix not found ~s~n" name)]	
       [(not svar)
	(error 'add-matrix-pragma "stride variable not found ~s~n" stride)]
       [else
	(set-memory_region-matrix_info!
	 mregion
	 (cons
	  (make-one_matrix
	   mvar
	   `(var ,svar)
	   '()
	   '()
	   '())
	    (memory_region-matrix_info mregion)))]))))
  
(define add-index-pragma
  (lambda (scil name type func)
    (let ([mregion (find-func scil func)])
      (if (not mregion)
          (error 'add-index-pragma "function not found ~s~n" func)
          #t))))    
      
(define add-pragmas
  (lambda (ps scil)
    (match ps
      [() scil]
      [((matrix-pragma ,name ,stride ,func) . ,rest)
       (add-matrix-pragma scil name stride func)
       (add-pragmas rest scil)]
      [((index-pragma ,name ,type ,func) . ,rest)
       (add-index-pragma scil name type func)
       (add-pragmas rest scil)]
      [(,otherwise . ,rest) (error 'add-pragmas "unrecognized pragma ~s~n" otherwise)])))      
              
(define add-pragma-data              
  (lambda (pair)
    (let ([scil (car pair)]
          [pragmas (cdr pair)])
       (add-pragmas pragmas scil)))) 
        
(define print-sig
  (lambda (routine)
    (printf "  ~s~n" (cons (routine-unmangled_name routine) (routine-param_typing routine)))))

(define print-sigs-wrap
  (lambda (rlist)
    (printf "(~n")
    (map print-sig rlist)
    (printf ")~n")))

(define output-func-signatures
  (lambda (scil-and-file)
    (let* ([scil (car scil-and-file)]
	   [file (cdr scil-and-file)]
	   [routines (memory_region-routine_list (scil_structure-file_scope scil))])
      (with-output-to-file file (lambda () (print-sigs-wrap routines)) 'replace))))

(define output-final-c-files 
 (lambda (scil-and-file)
    (let* ([scil (car scil-and-file)]
	   [file (cdr scil-and-file)])
      (with-output-to-file file (lambda () (display (handle-scil_structure scil))) 'replace))))

(define opie-compile-all
  (lambda (files)
    (let* ([scils (map build-tree-from-file (add-extension ".scil" files))]
           [pragmas (map read-pragma-data (add-extension ".pragmas" files))]
           [scils-with-pragmas (map add-pragma-data (zip scils pragmas))]
	   
           [compiled-scils (map morton-order-ss scils-with-pragmas)]
           [finalized-scils (map (update-function-calls compiled-scils) compiled-scils)])
      ;(printf "bah")
      (map output-func-signatures (zip compiled-scils (add-extension ".sig" files)))
      (map output-final-c-files (zip finalized-scils (add-extension ".morton.c" files)))
      '())))
;      compiled-scils))) 




