;; -----------------------------------------------------------------------
 ; Indiana University Opie Lab Software License, Version 1.2
 ;
 ; Copyright (C) 2002 The Trustees of Indiana University.
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without
 ; modification, are permitted provided that the following conditions are
 ; met:
 ;
 ; 1) All redistributions of source code must retain the above
 ;    copyright notice, the list of authors in the original source
 ;    code, this list of conditions and the disclaimer listed in this
 ;    license;
 ;
 ; 2) All redistributions in binary form must reproduce the above
 ;    copyright notice, this list of conditions and the disclaimer
 ;    listed in this license in the documentation and/or other
 ;    materials provided with the distribution;
 ;
 ; 3) Any documentation included with all redistributions must include
 ;    the following acknowledgement:
 ;
 ;      "This product includes software developed by the Indiana
 ;      University Opie Lab.  For further information please visit
 ;      http://www.cs.indiana.edu/~dswise"
 ;
 ;    Alternatively, this acknowledgment may appear in the software
 ;    itself, and wherever such third-party acknowledgments normally
 ;    appear.
 ;
 ; 4) The name "Indiana University" or "Indiana University
 ;    Opie Lab" shall not be used to endorse or promote
 ;    products derived from this software without prior written
 ;    permission from Indiana University.  For written permission,
 ;    please contact http://www.cs.indiana.edu/~dswise.
 ;
 ; 5) Products derived from this software may not use "Indiana
 ;    Univeristy" name nor may "Indiana University" appear in their name,
 ;    without prior written permission of the Indiana University.
 ;
 ; Indiana University provides no reassurances that the source code
 ; provided does not infringe the patent or any other intellectual
 ; property rights of any other entity.  Indiana University disclaims any
 ; liability to any recipient for claims brought by any other entity
 ; based on infringement of intellectual property rights or otherwise.
 ;
 ; LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH
 ; NO WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA
 ; UNIVERSITY GIVES NO WARRANTIES AND MAKES NO REPRESENTATION THAT
 ; SOFTWARE IS FREE OF INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR
 ; OTHER PROPRIETARY RIGHTS.  INDIANA UNIVERSITY MAKES NO WARRANTIES THAT
 ; SOFTWARE IS FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP
 ; DOORS", "WORMS", OR OTHER HARMFUL CODE.  LICENSEE ASSUMES THE ENTIRE
 ; RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR ASSOCIATED MATERIALS,
 ; AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION GENERATED USING
 ; SOFTWARE.
;; ------------------------------------------------------------------------

;
; Util function.  Returns the nth function memory region from the scil
; structure ss.
;
(define get-fmr
  (lambda (ss n) 
    (list-ref (scil_structure-fun_mem_region_list ss) n)
  )
)



; a-r :: (addr a) -> a
;
; Shorthand for address-ref which gets used a lot.
;
; This also allows the protocol of having any non-
; address (in practice () or 0) behave as though it
; is a null address.
;
; The address record is kept around only to maintain
; compatibility with addresses as used by the
; scil->C compiler, otherwise I would happily do
; away with it.
;
(define a-r
  (lambda (addr)
    (if (not (address? addr))
	'()
	(address-ref addr))))
;
; Ignore casts and just return the expression that is left over.
;
; strip-casts :: expr -> expr
(define strip-casts
  (lambda (expr)
    (cond
     [(null? expr) (error 'strip-casts "null expr sent to strip-casts~n")]
     [(equal? (expression-operation_kind expr) 'eok_cast)
      (strip-casts (a-r (expression-operands expr)))
     ]
     [else expr]
    )
  )
)

; strict-strip-casts :: expr -> expr|error
(define strict-strip-casts
  (lambda (expr)
    (cond
     [(null? expr) (error 'strip-casts "null expr sent to strip-casts~n")]
     [(equal? (expression-operation_kind expr) 'eok_cast)
      (strip-casts (a-r (expression-operands expr)))]
     [else (if (or
                 (equal?(expression-kind expr) 'enk_variable)
                 (equal?(expression-kind expr) 'enk_constant))
               expr 
               (error 'strict-strip-casts "casting of nonleaf expression~n"))])))

(define find-last-stmt-in-list
  (lambda (stmt)
    (cond
     [(nil? stmt) (error 'find-last-statement-in-list "must receive non-null argument")]
     [(nil? (a-r (statement-next stmt))) stmt]
     [else (find-last-stmt-in-list (a-r (statement-next stmt)))])))

(define find-last-stmt-in-block
  (lambda (stmt)
    ;(printf "entered find-last-stmt-in-block with: ~s~n" stmt)
    (cond
     [(or (nil? stmt) (not (statement? stmt)) (not (equal? (statement-kind stmt) 'stmk_block)))
      (error 'find-last-stmt-in-block "must receive a block statement as an argument")]
     [else (find-last-stmt-in-list (a-r (statement-statements stmt)))])))

(define for-with-initialization?
  (lambda (stmt)
    (if (and (statement? stmt) 
	     (equal? (statement-kind stmt) `stmk_for) 
	     (not (nil? (statement-initialization stmt))))
	#t
	#f)))

(define for-with-nontrivial-test?
  (lambda (stmt)
    (let ([test (a-r (statement-expr stmt))])
      (if (and (statement? stmt) 
	       (equal? (statement-kind stmt) 'stmk_for) 
	       (not (nil? test))
	       (not (equal? (expression-operation_kind test) `eok_my_one_const)))
	  #t
	  #f))))

(define for-with-increment?
  (lambda (stmt)
    (if (and (statement? stmt) 
	     (equal? (statement-kind stmt) ' stmk_for) 
	     (not (nil? (statement-increment stmt))))
	#t
	#f)))

(define for-with-question-initialization?
  (lambda (stmt)
    (let* ([init (a-r (statement-initialization stmt))]
	   [init-exprs (stmt->exprs init)])
      (if (and (statement? stmt) 
	       (equal? (statement-kind stmt) `stmk_for) 
	       (not (nil? (statement-initialization stmt)))
	       (ormap (lambda (x) x) (map (lambda (y) (equal? (expression-kind y) `enk_question)) init-exprs)))
	  #t
	  #f))))

(define for-with-question-test?
  (lambda (stmt)
    (let* ([test (a-r (statement-expr stmt))]
	   [test-exprs (stmt->exprs test)])
      (if (and (statement? stmt) 
	       (equal? (statement-kind stmt) 'stmk_for) 
	       (not (nil? test))
	       (ormap (lambda (x) x) (map (lambda (y) (equal? (expression-kind y) `enk_question)) test-exprs)))
	  #t
	  #f))))

(define for-with-question-increment?
  (lambda (stmt)
    (let* ([incr (statement-increment stmt)]
	   [incr-exprs (stmt->exprs incr)])
    (if (and (statement? stmt) 
	     (equal? (statement-kind stmt) ' stmk_for) 
	     (not (nil? (statement-increment stmt)))
	     (ormap (lambda (x) x) (map (lambda (y) (equal? (expression-kind y) `enk_question)) incr-exprs)))
	#t
	#f))))


;*********************************************************************************************
; CHANGE MADE HERE -- mrcarlso 6/04
;*********************************************************************************************
(define add-as-first-dimension
  (lambda (elt lst)
    (letrec ([helper
	   (lambda (elt lst accum)
	     (cond
	      [(null? lst) (append accum elt)]
	      [(equal? (car lst) "[") (append accum elt lst)]
	      [else (helper elt (cdr lst) (reverse (cons (car lst) (reverse accum))))]))])
      (helper elt lst '()))))
;********************************************************************************************

;*********************************************************************************************
; CHANGE MADE HERE -- mrcarlso 6/04
;*********************************************************************************************
(define get-type-before-brackets
  (lambda (ls)
    (letrec ([helper
	   (lambda (ls accum)
	     (cond
	      [(null? ls) accum]
	      [(equal? (car ls) "[") accum]
	      [else (helper (cdr ls) (reverse (cons (car ls) (reverse accum))))]))])
      (helper ls '()))))
;*********************************************************************************************
